#!/usr/bin/env python3
"""
TCRD db client utility (see also Pharos GraphQL API)
"""
import os,sys,argparse,re,time,json,logging

from ...idg import tcrd
from ...util import yaml as util_yaml
from ...util import db as util_db

#############################################################################
if __name__=='__main__':
  PARAM_FILE = os.environ['HOME']+"/.tcrd.yaml"
  idtypes=['TID', 'GENEID', 'UNIPROT', 'GENESYMB', 'ENSP']
  epilog = f"default param_file: {PARAM_FILE}"
  parser = argparse.ArgumentParser(description='TCRD MySql client utility', epilog=epilog)
  ops = ['info', 'listTables', 'listColumns', 'tableRowCounts', 'tdlCounts', 
	'listTargets', 'listXrefTypes', 'listXrefs', 'listDatasets',
	'listTargetsByDTO', 'listTargetFamilies',
	'listPhenotypes', 'listPhenotypeTypes',
	'listPublications',
	'listCompounds', 'listDrugs',
	'getTargets', 'getTargetsByXref', 'getTargetPage',
	'listDiseases', 'listDiseaseTypes',
	'getDiseaseAssociations', 'getDiseaseAssociationsPage',
	'getTargetpathways']
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--i", dest="ifile", help="input target ID file")
  parser.add_argument("--ids", help="input IDs")
  parser.add_argument("--idtype", choices=idtypes, default='TID', help='target ID type')
  parser.add_argument("--xreftypes", help='Xref types, comma-separated')
  parser.add_argument("--tdls", help="TDLs, comma-separated ({})".format('|'.join(tcrd.TDLS)))
  parser.add_argument("--tfams", help="target families, comma-separated")
  parser.add_argument("--param_file", default=PARAM_FILE)
  parser.add_argument("--dbhost")
  parser.add_argument("--dbport")
  parser.add_argument("--dbusr")
  parser.add_argument("--dbpw")
  parser.add_argument("--dbname")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  parser.add_argument("-q", "--quiet", action="store_true", help="Suppress progress notification.")
  args = parser.parse_args()

  # logging.PROGRESS = 15 (custom)
  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>0 else logging.ERROR if args.quiet else 15))

  params = util_yaml.ReadParamFile(args.param_file) if os.path.isfile(args.param_file) else {}
  if args.dbhost: params['DBHOST'] = args.dbhost 
  if args.dbport: params['DBPORT'] = args.dbport 
  if args.dbusr: params['DBUSR'] = args.dbusr 
  if args.dbpw: params['DBPW'] = args.dbpw
  if args.dbname: params['DBNAME'] = args.dbname 

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    logging.info('Input IDs: %d'%(len(ids)))
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)

#  try:
#    import mysql.connector as mysql
#    dbcon = mysql.connect(host=params['DBHOST'], port=params['DBPORT'], user=params['DBUSR'], passwd=params['DBPW'], db=params['DBNAME'])
#  except Exception as e:
#    logging.error(f'{e}')
#    try:
#      import MySQLdb as mysql
#      dbcon = mysql.connect(host=params['DBHOST'], port=int(params['DBPORT']), user=params['DBUSR'], passwd=params['DBPW'], db=params['DBNAME'])
#    except Exception as e2:
#      logging.error(f'{e2}')
#      sys.exit(1)

  try:
    dbcon = util_db.MySqlConnect(dbhost=params['DBHOST'], dbport=params['DBPORT'], dbusr=params['DBUSR'], dbpw=params['DBPW'], dbname=params['DBNAME'])
  except Exception as e:
    logging.error(f'{e}')
    sys.exit(1)

  if args.op=='listColumns':
    tcrd.Utils.ListColumns(dbcon, fout)

  elif args.op=='info':
    tcrd.Utils.Info(dbcon, fout)

  elif args.op=='tableRowCounts':
    tcrd.Utils.TableRowCounts(dbcon, fout)

  elif args.op=='tdlCounts':
    tcrd.Utils.TDLCounts(dbcon, fout)

  elif args.op=='listTables':
    tcrd.Utils.ListTables(dbcon, fout)

  elif args.op=='listTargets':
    tdls = re.split(r'\s*,\s*', args.tdls) if args.tdls else []
    tfams = re.split(r'\s*,\s*', args.tfams) if args.tfams else []
    tcrd.Utils.ListTargets(dbcon, tdls, tfams, fout)

  elif args.op=='getTargets':
    if not ids:
      parser.error(f'IDs required for operation: {args.op}')
    tcrd.Utils.GetTargets(dbcon, ids, args.idtype, fout)

  elif args.op=='getTargetPage':
    if not ids:
      parser.error(f'Target ID required for operation: {args.op}')
    tcrd.Utils.GetTargetPage(dbcon, ids[0], fout)

  elif args.op=='getTargetsByXrefs':
    if not ids:
      parser.error(f'IDs required for operation: {args.op}')
    tcrd.Utils.GetTargetsByXrefs(dbcon, ids, args.xreftypes, fout)

  elif args.op=='getTargetpathways':
    if not ids:
      parser.error(f'IDs required for operation: {args.op}')
    tids = tcrd.Utils.GetTargets(dbcon, ids, args.idtype, None)
    tcrd.Utils.GetPathways(dbcon, tids, fout)

  elif args.op=='listXrefTypes':
    tcrd.Utils.ListXrefTypes(dbcon, fout)

  elif args.op=='listXrefs':
    if args.xreftypes:
      xreftypes = re.split(r'\s*,\s*', args.xreftypes.strip())
      xreftypes_all = tcrd.Utils.ListXrefTypes(dbcon).iloc[:,0]
      for xreftype in xreftypes:
        if xreftype not in list(xreftypes_all):
          parser.error(f"xreftype '{xreftype}' invalid.  Available xreftypes: {str(list(xreftypes_all))}")
    else:
      xreftypes = []
    tcrd.Utils.ListXrefs(dbcon, xreftypes, fout)

  elif args.op=='listTargetFamilies':
    tcrd.Utils.ListTargetFamilies(dbcon, fout)

  elif args.op=='listTargetsByDTO':
    tcrd.Utils.ListTargetsByDTO(dbcon, fout)

  elif args.op=='listDiseases':
    tcrd.Utils.ListDiseases(dbcon, fout)

  elif args.op=='listDiseaseTypes':
    tcrd.Utils.ListDiseaseTypes(dbcon, fout)

  elif args.op=='getDiseaseAssociations':
    if not ids:
      parser.error(f'IDs required for operation: {args.op}')
    tcrd.Utils.GetDiseaseAssociations(dbcon, ids, fout)

  elif args.op=='getDiseaseAssociationsPage':
    if not ids:
      parser.error(f'ID required for operation: {args.op}')
    tcrd.Utils.GetDiseaseAssociationsPage(dbcon, ids[0], fout)

  elif args.op=='listPhenotypes':
    tcrd.Utils.ListPhenotypes(dbcon, fout)

  elif args.op=='listPhenotypeTypes':
    tcrd.Utils.ListPhenotypeTypes(dbcon, fout)

  elif args.op=='listPublications':
    tcrd.Utils.ListPublications(dbcon, fout)

  elif args.op=='listDatasets':
    tcrd.Utils.ListDatasets(dbcon, fout)

  elif args.op=='listCompounds':
    tcrd.Utils.ListCompounds(dbcon, fout)

  elif args.op=='listDrugs':
    tcrd.Utils.ListDrugs(dbcon, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")
    parser.print_help()



#!/usr/bin/env python3
"""
BRN = Biochemical Network Analysis (pybrn)
"""
import sys,os,re,logging
import numpy

from .. import reactome

try:
  import brn
except Exception as e:
  logging.error("pybrn not installed.")
  sys.exit()
#
API_HOST='reactomews.oicr.on.ca:8080'
BASE_PATH='/ReactomeRESTfulAPI/RESTfulWS'
API_BASE_URL='http://'+API_HOST+BASE_PATH
#

net = brn.fromSBML("data/reactome_reactions_homo_sapiens.2.sbml")

logging.info('reactions: %d'%len(net.reactions))
logging.info('species: %d'%len(net.species))
logging.info('values: %d'%len(net.values))

n_reac=0;
for r in net.reactions:
  n_reac+=1
  logging.info('%3d. %s'%(n_reac,net.showreact(r,printstr=False)))
logging.info('n_reac: %d'%n_reac)


n_spec=0;
for s in net.species:
  n_spec+=1
  id_spec = re.sub(r'[^\d]','',s)
  spec = reactome.Utils.GetId(API_BASE_URL,id_spec,'PhysicalEntity')
  displayName = spec['displayName'] if 'displayName' in spec else ''
  schemaClass = spec['schemaClass'] if 'schemaClass' in spec else ''
  logging.info('%3d. %s: (%s) %s'%(n_spec,id_spec,schemaClass,displayName))
logging.info('n_spec: %d'%n_spec)



#!/usr/bin/env python3
##############################################################################
### See http://www.guidetopharmacology.org/webServices.jsp
##############################################################################
import sys,os,re,time,logging,urllib.parse,json
#
from ..util import rest
#
INT_TAGS=[
	"interactionId",
	"targetId",
	"ligandAsTargetId",
	"targetSpecies",
	"primaryTarget",
	"targetBindingSite",
	"ligandId",
	"ligandContext",
	"endogenous",
	"type",
	"action",
	"actionComment",
	"selectivity",
	"concentrationRange",
	"affinity",
	"affinityType",
	"originalAffinity",
	"originalAffinityType",
	"originalAffinityRelation",
	"assayDescription",
	"assayConditions",
	"useDependent",
	"voltageDependent",
	"voltage",
	"physiologicalVoltage"
]
INT_REF_TAGS=[
	"referenceId",
	"pmid",
	"type",
	"title",
	"year",
	"volume",
	"issue",
	"pages",
	"articleTitle",
	"authors",
	"editors",
	"publisher",
	"publisherAddress",
	"pubStatus",
	"website",
	"doi",
	"url",
	"isbn",
	"patentNumber",
	"assignee",
	"accessDate",
	"modifyDate",
	"priorityDate",
	"publicationDate"
]
#############################################################################
def ListTargets(base_url, tgt_type, db, ids, fout):
  n_tgt=0; n_out=0; n_err=0;
  tags=None;
  url=base_url+'/targets?'
  if tgt_type: url+=('&type=%s'%urllib.parse.quote(tgt_type))
  if db: url+=('&database=%s'%(db))
  for id_this in ids:
    url_this = url+('&accession=%s'%(urllib.parse.quote(id_this)))
    rval=rest.GetURL(url_this, parse_json=True)
    tgts=rval
    for tgt in tgts:
      n_tgt+=1
      if n_tgt==1 or not tags:
        tags=sorted(tgt.keys())
        fout.write(('\t'.join(tags))+'\n')
      vals=[]
      for tag in tags:
        vals.append((tgt[tag]) if tag in tgt else '')
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
    
  #logging.info(json.dumps(rval, indent=2))
  logging.info('n_tgt: %d'%n_tgt)
  logging.info('n_out: %d'%n_out)

############################# ################################################
def ListTargetFamilies(base_url,  tgt_type, fout):
  n_fam=0; n_out=0; n_err=0;
  tags=None;
  url=base_url+'/targets/families'
  if tgt_type: url+=('?type=%s'%urllib.parse.quote(tgt_type))
  rval=rest.GetURL(url, parse_json=True)
  fams=rval
  for fam in fams:
    n_fam+=1
    if n_fam==1 or not tags:
      tags=sorted(fam.keys())
      fout.write(('\t'.join(tags))+'\n')
    vals=[]
    for tag in tags:
      vals.append((fam[tag]) if tag in fam else '')
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1
    
  #logging.info(json.dumps(rval, indent=2))
  logging.info('n_fam: %d'%n_fam)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetTarget(base_url, ids, fout):
  n_qry=0; n_tgt=0; n_out=0; n_err=0;
  tags=None;
  for tid in ids:
    n_qry+=1
    url=base_url+'/targets/%s'%tid
    rval=rest.GetURL(url, parse_json=True)
    tgt=rval
    n_tgt+=1
    if n_tgt==1 or not tags:
      tags=sorted(tgt.keys())
      fout.write(('\t'.join(tags))+'\n')
    vals=[]
    for tag in tags:
      vals.append((tgt[tag]) if tag in tgt else '')
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1

  logging.info('n_tgt: %d'%n_tgt)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetTargetSubstrates(base_url, ids, species, fout):
  n_qry=0; n_tsb=0; n_out=0; n_err=0;
  tags=None;
  for id_query in ids:
    n_qry+=1
    url=base_url+('/targets/%s/substrates'%id_query)
    if species: url+=('?species=%s'%urllib.parse.quote(species))
    rval=rest.GetURL(url, parse_json=True)
    tsbs=rval
    for tsb in tsbs:
      n_tsb+=1
      if n_tsb==1 or not tags:
        tags=sorted(tsb.keys())
        fout.write(('\t'.join(tags))+'\n')
      vals=[]
      for tag in tags:
        vals.append((tsb[tag]) if tag in tsb else '')
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1

  logging.info('n_tsb: %d'%n_tsb)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetTargetProducts(base_url, ids, species, fout):
  n_qry=0; n_prd=0; n_out=0; n_err=0;
  tags=None;
  for id_query in ids:
    n_qry+=1
    url=base_url+('/targets/%s/products'%id_query)
    if species: url+=('?species=%s'%urllib.parse.quote(species))
    rval=rest.GetURL(url, parse_json=True)
    prds=rval
    for prd in prds:
      n_prd+=1
      if n_prd==1 or not tags:
        tags=sorted(prd.keys())
        fout.write(('\t'.join(tags))+'\n')
      vals=[]
      for tag in tags:
        vals.append((prd[tag]) if tag in prd else '')
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1

  logging.info('n_prd: %d'%n_prd)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetTargetFunction(base_url, ids, species, fout):
  n_qry=0; n_fnc=0; n_out=0; n_err=0;
  tags=None;
  for id_query in ids:
    n_qry+=1
    url=base_url+('/targets/%s/function'%id_query)
    if species: url+=('?species=%s'%urllib.parse.quote(species))
    rval=rest.GetURL(url, parse_json=True)
    fncs=rval
    for fnc in fncs:
      n_fnc+=1
      if n_fnc==1 or not tags:
        tags=sorted(fnc.keys())
        fout.write(('\t'.join(tags))+'\n')
      vals=[]
      for tag in tags:
        vals.append((fnc[tag]) if tag in fnc else '')
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1

  logging.info('n_fnc: %d'%n_fnc)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetTargetGeneProteinInfo(base_url, ids, fout):
  n_qry=0; n_gpi=0; n_out=0; n_err=0;
  tags=None;
  for tid in ids:
    n_qry+=1
    url=base_url+'/targets/%s/geneProteinInformation'%tid
    rval=rest.GetURL(url, parse_json=True)
    gpis=rval
    for gpi in gpis:
      n_gpi+=1
      if n_gpi==1 or not tags:
        tags=sorted(gpi.keys())
        fout.write(('\t'.join(tags))+'\n')
      vals=[]
      for tag in tags:
        vals.append((gpi[tag]) if tag in gpi else '')
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1

  logging.info('n_gpi: %d'%n_gpi)
  logging.info('n_out: %d'%n_out)

#############################################################################
def ListLigands(base_url, lig_type, db, ids, fout):
  n_lig=0; n_out=0; n_err=0;
  tags=None;
  url=base_url+'/ligands?'
  if lig_type: url+=('&type=%s'%urllib.parse.quote(lig_type))
  if db: url+=('&database=%s'%(db))
  for id_this in ids:
    url_this = url+('&database=%s&accession=%s'%(db, urllib.parse.quote(id_this)))
    rval = rest.GetURL(url_this, parse_json=True)
    ligs=rval
    for lig in ligs:
      n_lig+=1
      if n_lig==1 or not tags:
        tags=sorted(lig.keys())
        fout.write(('\t'.join(tags))+'\n')
      vals=[]
      for tag in tags:
        vals.append((lig[tag]) if tag in lig else '')
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
    
  logging.info('n_lig: %d'%n_lig)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetLigand(base_url, ids, fout):
  n_qry=0; n_lig=0; n_out=0; n_err=0;
  tags=None;
  for lid in ids:
    n_qry+=1
    url=base_url+'/ligands/%s'%lid
    rval=rest.GetURL(url, parse_json=True)
    lig=rval
    n_lig+=1
    if n_lig==1 or not tags:
      tags=sorted(lig.keys())
      fout.write(('\t'.join(tags))+'\n')
    vals=[]
    for tag in tags:
      vals.append((lig[tag]) if tag in lig else '')
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1

  logging.info('n_lig: %d'%n_lig)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetLigandStructure(base_url, ids, fout):
  '''Smiles backslashes may cause problems.'''
  n_qry=0; n_lig=0; n_out=0; n_err=0;
  tags=None;
  for lid in ids:
    n_qry+=1
    url=base_url+'/ligands/%s/structure'%lid
    rval=rest.GetURL(url, parse_json=True)
    lig=rval
    if not lig:
      logging.error('Not found: %s'%(url))
      n_err+=1
      continue
    n_lig+=1
    if n_lig==1 or not tags:
      tags=sorted(lig.keys())
      fout.write(('\t'.join(tags))+'\n')
    vals=[]
    for tag in tags:
      vals.append((lig[tag]) if tag in lig else '')
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1

  logging.info('n_lig: %d'%n_lig)
  logging.info('n_out: %d'%n_out)
  logging.info('n_err: %d'%n_err)

#############################################################################
def GetInteractions(base_url, resource, ids, species, itr_type, aff_type, aff, fout):
  '''Include first reference only.'''
  n_qry=0; n_itr=0; n_out=0; n_err=0;
  n_dpt=0;
  for id_query in ids:
    n_qry+=1
    url=base_url+('/%s/%s/interactions?'%(resource, id_query))
    if species: url+=('&species=%s'%urllib.parse.quote(species))
    if itr_type: url+=('&type=%s'%urllib.parse.quote(itr_type))
    if aff_type: url+=('&affinityType=%s'%urllib.parse.quote(aff_type))
    if aff: url+=('&affinity=%.2f'%aff)
    rval=rest.GetURL(url, parse_json=True)
    itrs=rval
    if not itrs or type(itrs) not in (list, tuple):
      logging.error('Not found: %s'%(url))
      n_err+=1
      continue
    for itr in itrs:
      n_itr+=1
      dpts = itr['dataPoints'] if 'dataPoints' in itr else []
      for dpt in dpts:
        n_dpt+=1
        if n_dpt==1:
          fout.write(('\t'.join(INT_TAGS+INT_REF_TAGS))+'\n')
        vals=[]
        for tag in INT_TAGS:
          vals.append((dpt[tag]) if tag in dpt else '')
        refs = dpt['refs'] if 'refs' in dpt else []
        if refs:
          for tag in INT_REF_TAGS:
            vals.append((refs[0][tag]) if tag in refs[0] else '')
        else:
          vals.extend(['' for tag in INT_REF_TAGS])
        fout.write(('\t'.join(vals))+'\n')
        n_out+=1

  logging.info('n_itr: %d'%n_itr)
  logging.info('n_dpt: %d'%n_dpt)
  logging.info('n_out: %d'%n_out)
  logging.info('n_err: %d'%n_err)

#############################################################################
def SearchLigand(base_url, smi, search_type, fout):
  n_lig=0; n_out=0; n_err=0;
  tags=None;
  url=base_url+('/ligands/%s?smiles=%s'%(search_type, urllib.parse.quote(smi)))
  rval=rest.GetURL(url, parse_json=True)
  ligs=rval
  for lig in ligs:
    n_lig+=1
    if n_lig==1 or not tags:
      tags=sorted(lig.keys())
      fout.write(('\t'.join(tags))+'\n')
    vals=[]
    for tag in tags:
      vals.append((lig[tag]) if tag in lig else '')
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1
    
  logging.info('n_lig: %d'%n_lig)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetSynonyms(base_url, resource, ids, fout):
  '''resource = targets|ligands'''
  n_qry=0; n_syn=0; n_out=0; n_err=0;
  tags=None;
  for id_query in ids:
    n_qry+=1
    url=base_url+'/%s/%s/synonyms'%(resource, id_query)
    rval=rest.GetURL(url, parse_json=True)
    syns=rval
    for syn in syns:
      n_syn+=1
      if n_syn==1 or not tags:
        tags=sorted(syn.keys())
        fout.write(('\t'.join(['%s_id'%resource]+tags))+'\n')
      vals=[]
      for tag in tags:
        vals.append((syn[tag]) if tag in syn else '')
      fout.write(('\t'.join([id_query]+vals))+'\n')
      n_out+=1

  logging.info('n_syn: %d'%n_syn)
  logging.info('n_out: %d'%n_out)

#############################################################################
def GetDblinks(base_url, resource, ids, species, db, fout):
  n_qry=0; n_dbl=0; n_out=0; n_err=0;
  tags=None;
  for id_query in ids:
    n_qry+=1
    url=base_url+'/%s/%s/databaseLinks?'%(resource, id_query)
    if species: url+=('&species=%s'%urllib.parse.quote(species))
    if db: url+=('&database=%s'%urllib.parse.quote(db))
    rval=rest.GetURL(url, parse_json=True)
    dbls=rval
    for dbl in dbls:
      n_dbl+=1
      if n_dbl==1 or not tags:
        tags=sorted(dbl.keys())
        fout.write(('\t'.join(['%s_id'%resource]+tags))+'\n')
      vals=[]
      for tag in tags:
        vals.append((dbl[tag]) if tag in dbl else '')
      fout.write(('\t'.join([id_query]+vals))+'\n')
      n_out+=1

  logging.info('n_dbl: %d'%n_dbl)
  logging.info('n_out: %d'%n_out)

#############################################################################



#!/usr/bin/env python3
#############################################################################
### For accessing files via FTP site; ftp://ftp.ncbi.nlm.nih.gov/pubchem/
#############################################################################
import sys,os,re,time,argparse,logging

from ... import pubchem

FTP_URL='ftp://ftp.ncbi.nlm.nih.gov/pubchem'
POLL_WAIT=10
MAX_WAIT=600

#############################################################################
if __name__=='__main__':
  epilog="""
FTP_URL: {0}""".format(FTP_URL)
  parser = argparse.ArgumentParser(description="access PubChem FTP site", epilog=epilog)
  parser.add_argument("--ftp_get", help="path of file")
  parser.add_argument("--ftp_ls", help="path of dir")
  parser.add_argument("--ftp_url", default=FTP_URL)
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--nmax", type=int, default=0)
  parser.add_argument("--ftp_ntries", type=int, default=20, help="max tries per ftp-get")
  parser.add_argument("--sdf2smi", action="store_true", help="convert SDF to SMILES")
  parser.add_argument("--o", dest="ofile", help="output file")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  if not (args.ftp_get or args.ftp_ls):
    parser.error("--ftp_get or --ftp_ls required.")

  if args.ofile:
    fout = open(args.ofile, "w")
  else:
    fout = sys.stdout

  if args.ftp_get:
    url=("%s%s"%(args.ftp_url, args.ftp_get))
    if args.sdf2smi:
      nbytes = pubchem.ftp.Utils.GetUrlSDF2SMI(url, fout, ntries=args.ftp_ntries, poll_wait=10)
    else:
      nbytes = pubchem.ftp.Utils.GetUrl(url, fout, ntries=args.ftp_ntries, poll_wait=10)
    logging.info("bytes: %.2fMB"%(nbytes/1e6))
  elif args.ftp_ls:
    url=("%s%s"%(args.ftp_url, args.ftp_ls))
    pubchem.ftp.Utils.GetUrl(url, fout, ntries=args.ftp_ntries, poll_wait=10)



#!/usr/bin/env python3
"""
Utility functions for ChEMBL REST API.
https://chembl.gitbook.io/chembl-interface-documentation/web-services/chembl-data-web-services
http://chembl.blogspot.com/2015/02/using-new-chembl-web-services.html
"""
###
import sys,os,re,json,time,urllib.parse,logging,tqdm
import pandas as pd
#
from ..util import rest
#
NCHUNK=100
#
API_HOST="www.ebi.ac.uk"
API_BASE_PATH="/chembl/api/data"
BASE_URL="https://"+API_HOST+API_BASE_PATH
#
##############################################################################
def Status(base_url=BASE_URL, fout=None):
  rval = rest.GetURL(f"{base_url}/status.json", parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2))
  df = pd.DataFrame({tag:(rval[tag] if tag in rval else "") for tag in rval.keys()})
  if fout is not None: df.to_csv(fout, "\t", index=False)
  else: return df

#############################################################################
def GetTargetByUniprot(ids, base_url=BASE_URL, fout=None):
  n_out=0;
  ids_chembl = set()
  fout.write("UniprotId\ttarget_chembl_id\n")
  for uniprot in ids:
    id_chembl=None
    rval = rest.GetURL(f"{base_url}/target.json?target_components__accession={uniprot}", parse_json=True)
    targets = rval["targets"] if "targets" in rval else []
    for target in targets:
      id_chembl = target["target_chembl_id"]
      ids_chembl.add(id_chembl)
      fout.write(f"{uniprot}\t\{id_chembl}\n")
      n_out+=1
    if len(ids_chembl)>1:
      logging.info(f"Uniprot ambiguous: {uniprot}")
    for id_chembl in list(ids_chembl):
      logging.debug(f"Uniprot: {uniprot} -> ChEMBL: {id_chembl}")
  logging.info(f"n_out: {n_out}")

#############################################################################
def GetActivity(ids, resource, pmin, skip=0, nmax=None, api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  '''Get activity data and necessary references only, due to size concerns.  resource = assay|target|molecule.  Filter on pChEMBL value, standardized negative log molar half-max response activity.'''
  n_act=0; n_out=0; n_pval=0; n_pval_ok=0; tags=None; tq=None;
  for i,id_this in enumerate(ids):
    if i<skip: continue
    if not tq: tq = tqdm.tqdm(total=len(ids)-skip, unit=resource+"s")
    tq.update()
    url_next = (f"{api_base_path}/activity.json?{resource}_chembl_id={id_this}&limit={NCHUNK}")
    while True:
      rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
      if rval is None: break
      acts = rval["activities"] if "activities" in rval else []
      for act in acts:
        logging.debug(json.dumps(act, sort_keys=True, indent=2))
        n_act+=1
        if not tags:
          tags = list(act.keys())
          for tag in tags[:]:
            if type(act[tag]) in (dict, list, tuple):
              tags.remove(tag)
              logging.debug(f'Ignoring field ({type(act[tag])}): "{tag}"')
          fout.write('\t'.join(tags)+'\n')
        vals = [(str(act[tag]) if tag in act else '') for tag in tags]
        if pmin is not None:
          try:
            pval = float(act["pchembl_value"])
            n_pval+=1
            if pval >= pmin:
              n_pval_ok+=1
              fout.write('\t'.join(vals)+'\n')
              n_out+=1
              logging.debug(f"[{n_act}] pVal ok ({pval:4.1f} >= {pmin:4.1f})")
            else:
              logging.debug(f"[{n_act}] pVal low ({pval:4.1f} < {pmin:4.1f})")
          except:
            logging.debug(f"[{n_act}] pVal missing.")
        else:
          fout.write('\t'.join(vals)+'\n')
          n_out+=1
      total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
      url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
      if not url_next: break
    if nmax and i>=(nmax-skip): break
  if tq is not None: tq.close()
  logging.info(f"n_qry: {len(ids)}; n_act: {n_act}; n_out: {n_out}")
  if pmin is not None:
    logging.info(f"n_pval: {n_pval}; n_pval_ok: {n_pval_ok}; pVals missing: {n_act-n_pval}")

#############################################################################
def GetActivityProperties(ids, skip=0, nmax=None, base_url=BASE_URL, fout=None):
  n_out=0; tags=None;
  for i,id_this in enumerate(ids):
    if i<skip: continue
    act = rest.GetURL((f"{base_url}/activity/{id_this}.json"), parse_json=True)
    assay_chembl_id = act["assay_chembl_id"] if "assay_chembl_id" in act else ""
    molecule_chembl_id = act["molecule_chembl_id"] if "molecule_chembl_id" in act else ""
    props = act["activity_properties"] if "activity_properties" in act else []
    for prop in props:
      if not tags:
        tags = list(prop.keys())
        fout.write('\t'.join(["activity_id", "assay_chembl_id", "molecule_chembl_id"]+tags)+"\n")
      logging.debug(json.dumps(prop, sort_keys=True, indent=2))
      vals = [str(prop[tag]) if tag in prop else "" for tag in tags]
      fout.write(('\t'.join([id_this, assay_chembl_id, molecule_chembl_id]+vals))+'\n')
      n_out+=1
    if nmax and i>=(nmax-skip): break
  logging.info(f"n_qry: {len(ids)}; n_out: {n_out}")

#############################################################################
def ListTargets(skip, nmax, api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  '''One row per target.  Ignore synonyms. If one-component, single protein, include UniProt accession.'''
  n_tgt=0; n_cmt=0; n_out=0; tags=None; tq=None;
  url_next = (f"{api_base_path}/target.json?limit={NCHUNK}&offset={skip}")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    tgts = rval["targets"] if rval and "targets" in rval else []
    for tgt in tgts:
      logging.debug(json.dumps(tgt, indent=2))
      n_tgt+=1
      if not tags:
        tags = sorted(tgt.keys())
        for tag in tags[:]:
          if type(tgt[tag]) in (dict, list, tuple):
            tags.remove(tag)
            logging.debug(f'Ignoring field ({type(tgt[tag])}): "{tag}"')
        tags.extend(["component_count", "accession"])
        fout.write('\t'.join(tags)+'\n')
      vals = [str(tgt[tag]) if tag in tgt else "" for tag in tags]
      if "target_components" in tgt and tgt["target_components"]:
        cmts = tgt["target_components"]
        n_cmt+=len(cmts)
        vals.append(f"{len(cmts)}")
        vals.append(cmts[0]["accession"] if len(cmts)==1 else "")
      else:
        logging.debug(f"no-component target: {vals[0]}")
        vals.extend(["", ""])
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
      if tq is not None: tq.update()
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="tgts")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_targets: {n_tgt}; n_target_components: {n_cmt}; n_out: {n_out}")

#############################################################################
def GetTarget(ids, base_url=BASE_URL, fout=None):
  '''One row per target.  Ignore synonyms. If one-component, single protein, include UniProt accession.'''
  n_tgt=0; n_cmt=0; n_out=0; tags=None; tq=None;
  for id_this in ids:
    tgt = rest.GetURL(f"{base_url}/target/{id_this}.json", parse_json=True)
    if not tgt:
      logging.error(f'Not found: "{id_this}"')
      continue
    n_tgt+=1
    if not tq: tq = tqdm.tqdm(total=len(ids), unit="tgts")
    tq.update()
    if n_tgt==1 or not tags:
      tags = sorted(tgt.keys())
      for tag in tags[:]:
        if type(tgt[tag]) in (dict, list, tuple):
          tags.remove(tag)
          logging.debug(f'Ignoring field ({type(tgt[tag])}): "{tag}"')
      tags.extend(["component_count", "accession"])
      fout.write('\t'.join(tags)+'\n')
    logging.debug(json.dumps(tgt,sort_keys=True,indent=2))
    vals = [str(tgt[tag]) if tag in tgt else "" for tag in tags]
    if "target_components" in tgt and tgt["target_components"]:
      cmts = tgt["target_components"]
      n_cmt+=len(cmts)
      vals.append(f"{len(cmts)}")
      vals.append(str(cmts[0]["accession"]) if len(cmts)==1 else "")
    else:
      logging.debug(f"no-component target: {vals[0]}")
      vals.extend(["", ""])
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1
  if tq is not None: tq.close()
  logging.info(f"n_qry: {len(ids)}; n_targets: {n_tgt}; n_target_components: {n_cmt}; n_out: {n_out}")

#############################################################################
def GetTargetComponents(ids, skip=0, nmax=None, base_url=BASE_URL, fout=None):
  n_tgt=0; n_out=0; tags=[]; cmt_tags=[]; df=None; tq=None;
  for i,id_this in enumerate(ids):
    if i<skip: continue
    if not tq: tq = tqdm.tqdm(total=len(ids)-skip, unit="tgts")
    tq.update()
    tgt = rest.GetURL(f"{base_url}/target/{id_this}.json", parse_json=True)
    if not tgt: continue
    n_tgt+=1
    vals = [str(tgt[tag]) if tag in tgt else "" for tag in tags]
    cmts = tgt["target_components"] if "target_components" in tgt and tgt["target_components"] else []
    if not cmts: continue
    for cmt in cmts:
      logging.debug(json.dumps(cmt, indent=2))
      if not tags:
        for tag in tgt.keys():
          if type(tgt[tag]) not in (dict, list, tuple):
            tags.append(tag)
        for tag in cmt.keys():
          if type(cmt[tag]) not in (dict, list, tuple):
            cmt_tags.append(tag)
      df_this = pd.concat([
	pd.DataFrame({tag:[(tgt[tag] if tag in tgt else None)] for tag in tags}),
	pd.DataFrame({tag:[(cmt[tag] if tag in cmt else None)] for tag in cmt_tags})],
	axis=1)
      if fout is None:
        df = pd.concat([df, df_this])
      else:
        df_this.to_csv(fout, "\t", index=False)
      n_out+=df_this.shape[0]
    if nmax and i>=(nmax-skip): break
  if tq is not None: tq.close()
  logging.info(f"n_qry: {len(ids)}; n_targets: {n_tgt}; n_out: {n_out}")
  if fout is None: return df

#############################################################################
def GetDocument(ids, skip=0, nmax=None, base_url=BASE_URL, fout=None):
  n_pmid=0; n_doi=0; n_out=0; tags=None; tq=None;
  for i,id_this in enumerate(ids):
    if i<skip: continue
    if not tq: tq = tqdm.tqdm(total=len(ids)-skip, unit="docs")
    tq.update()
    doc = rest.GetURL(f"{base_url}/document/{id_this}.json", parse_json=True)
    if not doc:
      logging.error(f'Not found: "{id_this}"')
      continue
    if not tags:
      tags = list(doc.keys())
      fout.write('\t'.join(tags)+'\n')
    logging.debug(json.dumps(doc, sort_keys=True, indent=2))
    if "pubmed_id" in tags and doc["pubmed_id"]: n_pmid+=1
    if "doi" in tags and doc["doi"]: n_doi+=1
    vals = [str(doc[tag]) if tag in doc else "" for tag in tags]
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1
  if tq is not None: tq.close()
  logging.info(f"n_qry: {len(ids)}; n_pmid: {n_pmid}; n_doi: {n_doi}; n_out: {n_out}")

#############################################################################
def ListSources(api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_out=0; tags=None;
  url_next = (api_base_path+"/source.json")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    sources = rval["sources"] if "sources" in rval else []
    for source in sources:
      if not tags:
        tags = list(source.keys())
        fout.write('\t'.join(tags)+'\n')
      logging.debug(json.dumps(source, sort_keys=True, indent=2))
      vals = [str(source[tag]) if tag in source else "" for tag in tags]
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  logging.info(f"n_out: {n_out}")

#############################################################################
def ListCellLines(api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_clo=0; n_efo=0; n_out=0; tags=None;
  url_next = (api_base_path+"/cell_line.json")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    cells = rval["cell_lines"] if "cell_lines" in rval else []
    for cell in cells:
      if not tags:
        tags = list(cell.keys())
        fout.write('\t'.join(tags)+'\n')
      logging.debug(json.dumps(cell, sort_keys=True, indent=2))
      if "clo_id" in cell and cell["clo_id"]: n_clo+=1
      if "efo_id" in cell and cell["efo_id"]: n_efo+=1
      vals = [str(cell[tag]) if tag in cell else "" for tag in tags]
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  logging.info(f"n_out: {n_out}; n_clo: {n_clo}; n_efo: {n_efo}")

#############################################################################
def ListOrganisms(api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_out=0; tags=None;
  url_next = (api_base_path+"/organism.json")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    orgs = rval["organisms"] if "organisms" in rval else []
    for org in orgs:
      if not tags:
        tags = list(org.keys())
        fout.write('\t'.join(tags)+'\n')
      logging.debug(json.dumps(org, sort_keys=True, indent=2))
      vals = [str(org[tag]) if tag in org else "" for tag in tags]
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  logging.info(f"n_out: {n_out}")

#############################################################################
def ListProteinClasses(api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_out=0; tags=None;
  url_next = (api_base_path+"/protein_class.json")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    pcls = rval["protein_classes"] if "protein_classes" in rval else []
    for pcl in pcls:
      if not tags:
        tags = list(pcl.keys())
        fout.write('\t'.join(tags)+'\n')
      logging.debug(json.dumps(pcl, sort_keys=True, indent=2))
      vals = [str(pcl[tag]) if tag in pcl else "" for tag in tags]
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  logging.info(f"n_out: {n_out}")

#############################################################################
def ListDrugIndications(skip, nmax, api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_efo=0; n_out=0; tags=None; tq=None;
  url_next = (f"{api_base_path}/drug_indication.json?limit={NCHUNK}&offset={skip}")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    dins = rval["drug_indications"] if "drug_indications" in rval else []
    for din in dins:
      if not tags:
        tags = list(din.keys())
        for tag in tags[:]:
          if type(din[tag]) in (dict, list, tuple):
            tags.remove(tag)
          logging.debug(f'Ignoring field ({type(din[tag])}): "{tag}"')
        fout.write('\t'.join(tags)+'\n')
      logging.debug(json.dumps(din, sort_keys=True, indent=2))
      if "efo_id" in din and din["efo_id"]: n_efo+=1
      vals = [str(din[tag]) if tag in din else "" for tag in tags]
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
      if tq is not None: tq.update()
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="inds")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}; n_efo: {n_efo}")

#############################################################################
def ListTissues(api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_bto=0; n_efo=0; n_caloha=0; n_uberon=0; n_out=0; tags=None; tq=None;
  url_next = (api_base_path+"/tissue.json")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    tissues = rval["tissues"] if "tissues" in rval else []
    for tissue in tissues:
      if not tags:
        tags = list(tissue.keys())
        fout.write('\t'.join(tags)+'\n')
      logging.debug(json.dumps(tissue, sort_keys=True, indent=2))
      if "bto_id" in tissue and tissue["bto_id"]: n_bto+=1
      if "efo_id" in tissue and tissue["efo_id"]: n_efo+=1
      if "uberon_id" in tissue and tissue["uberon_id"]: n_uberon+=1
      if "caloha_id" in tissue and tissue["caloha_id"]: n_caloha+=1
      vals = [str(tissue[tag]) if tag in tissue else "" for tag in tags]
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
      if tq is not None: tq.update()
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="tissues")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}; n_bto: {n_bto}; n_efo: {n_efo}; n_caloha: {n_caloha}; n_uberon: {n_uberon}")

#############################################################################
def ListMechanisms(api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_out=0; tags=None; tq=None;
  url_next = (api_base_path+"/mechanism.json")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    mechs = rval["mechanisms"] if "mechanisms" in rval else []
    for mech in mechs:
      if not tags:
        tags = list(mech.keys())
        fout.write('\t'.join(tags)+'\n')
      logging.debug(json.dumps(mech, sort_keys=True, indent=2))
      vals = [str(mech[tag]) if tag in mech else "" for tag in tags]
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
      if tq is not None: tq.update()
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="mechs")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}")

#############################################################################
def ListDocuments(skip, nmax, api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_pmid=0; n_doi=0; n_out=0; n_err=0; tags=None; tq=None;
  url_next = (f"{api_base_path}/document.json?limit={NCHUNK}&offset={skip}")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    docs = rval["documents"] if "documents" in rval else []
    for doc in docs:
      if not tags:
        tags = list(doc.keys())
        if "abstract" in tags: tags.remove("abstract") #unnecessary, verbose
        fout.write('\t'.join(tags)+'\n')
      logging.debug(json.dumps(doc, sort_keys=True, indent=2))
      if "pubmed_id" in tags and doc["pubmed_id"]: n_pmid+=1
      if "doi" in tags and doc["doi"]: n_doi+=1
      vals = [str(doc[tag]) if tag in doc else "" for tag in tags]
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
      if tq is not None: tq.update()
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="docs")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}; n_pmid: {n_pmid}; n_doi: {n_doi}")

#############################################################################
def GetAssay(ids, base_url=BASE_URL, fout=None):
  n_out=0; tags=None;
  for id_this in ids:
    assay = rest.GetURL(f"{base_url}/assay/{id_this}.json", parse_json=True)
    if not assay:
      continue
    if not tags:
      tags = list(assay.keys())
      for tag in tags[:]:
        if type(assay[tag]) in (dict, list, tuple):
          tags.remove(tag)
          logging.debug(f'Ignoring field ({type(assay[tag])}): "{tag}"')
      fout.write('\t'.join(tags)+'\n')
    vals = [(str(assay[tag]) if tag in assay else "") for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info(f"n_in: {len(ids)}; n_out: {n_out}")

#############################################################################
def ListAssays(skip, nmax, api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_ass=0; n_out=0; tags=None; tq=None;
  url_next = (f"{api_base_path}/assay.json?offset={skip}&limit={NCHUNK}")
  t0 = time.time()
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    assays = rval["assays"] if "assays" in rval else []
    for assay in assays:
      n_ass+=1
      if not tags:
        tags = list(assay.keys())
        for tag in tags[:]:
          if type(assay[tag]) in (dict, list, tuple):
            tags.remove(tag)
            logging.debug(f'Ignoring field ({type(assay[tag])}): "{tag}"')
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(assay[tag]).replace('\t', " ") if tag in assay else "") for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
      if tq is not None: tq.update()
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="assays")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}")
  logging.info(f"""Elapsed time: {time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))}""")

#############################################################################
def SearchAssays(asrc, atype, skip, nmax, api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  '''Select assays based on source and optionally type.'''
  n_ass=0; n_out=0; tags=None; tq=None;
  url_next = (f"{api_base_path}/assay.json?offset={skip}&limit={NCHUNK}")
  if asrc: url_next+=(f"&src_id={asrc}")
  if atype: url_next+=(f"&assay_type={atype}")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    assays = rval["assays"] if "assays" in rval else []
    for assay in assays:
      n_ass+=1
      if not tags:
        tags = list(assay.keys())
        for tag in tags[:]:
          if type(assay[tag]) in (dict, list, tuple):
            tags.remove(tag)
            logging.debug(f'Ignoring field ({type(assay[tag])}): "{tag}"')
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(assay[tag]) if tag in assay else "") for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
      if tq is not None: tq.update()
      if nmax and n_out>=nmax: break
    if nmax and n_out>=nmax: break
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="assays")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_assay: {n_ass}; n_out: {n_out}")

##############################################################################
def GetMolecule(ids, base_url=BASE_URL, fout=None):
  '''Ignore molecule_synonyms.'''
  n_out=0; tags=None; struct_tags=None; prop_tags=None; tq=None;
  for id_this in ids:
    mol = rest.GetURL(f"{base_url}/molecule/{id_this}.json", parse_json=True)
    if not mol: continue
    if not tq: tq = tqdm.tqdm(total=len(ids), unit="mols")
    tq.update()
    if not tags:
      tags = sorted(list(mol.keys()))
      for tag in tags[:]:
        if type(mol[tag]) in (dict, list, tuple):
          tags.remove(tag)
          logging.debug(f'Ignoring field ({type(mol[tag])}): "{tag}"')
      struct_tags = sorted(mol["molecule_structures"].keys())
      struct_tags.remove("molfile")
      prop_tags = sorted(mol["molecule_properties"].keys())
      fout.write('\t'.join(tags+struct_tags+prop_tags+["parent_chembl_id"])+'\n')
    logging.debug(json.dumps(mol, sort_keys=True, indent=2))
    vals = [(mol["molecule_hierarchy"]["parent_chembl_id"] if "molecule_hierarchy" in mol and "parent_chembl_id" in mol["molecule_hierarchy"] else "")]
    vals.extend([(str(mol[tag]) if tag in mol else "") for tag in tags])
    vals.extend([(str(mol["molecule_structures"][tag]) if "molecule_structures" in mol and tag in mol["molecule_structures"] else "") for tag in struct_tags])
    vals.extend([(str(mol["molecule_properties"][tag]) if "molecule_properties" in mol and tag in mol["molecule_properties"] else "") for tag in prop_tags])
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1
  if tq is not None: tq.close()
  logging.info(f"n_in: {len(ids)}; n_out: {n_out}")

#############################################################################
def ListMolecules(dev_phase, skip, nmax, api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  '''Ignore synonyms here.'''
  n_mol=0; n_out=0; n_err=0; tags=None; struct_tags=None; prop_tags=None; tq=None;
  url_next=(f"{api_base_path}/molecule.json?limit={NCHUNK}")
  if skip: url_next+=(f"&offset={skip}")
  if dev_phase: url_next+=(f"&max_phase={dev_phase}")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    mols = rval["molecules"] if "molecules" in rval else []
    for mol in mols:
      n_mol+=1
      logging.debug(json.dumps(mol, sort_keys=True, indent=2))
      if not tags:
        tags = sorted(mol.keys())
        for tag in tags[:]:
          if type(mol[tag]) in (dict, list, tuple):
            tags.remove(tag)
            logging.debug(f'Ignoring field ({type(mol[tag])}): "{tag}"')
        struct_tags = sorted(mol["molecule_structures"].keys())
        struct_tags.remove("molfile")
        prop_tags = sorted(mol["molecule_properties"].keys())
        fout.write('\t'.join(tags+struct_tags+prop_tags+["parent_chembl_id"])+'\n')
      vals = [(mol["molecule_hierarchy"]["parent_chembl_id"] if "molecule_hierarchy" in mol and mol["molecule_hierarchy"] and "parent_chembl_id" in mol["molecule_hierarchy"] else "")]
      vals.extend([(str(mol[tag]) if tag in mol else "") for tag in tags])
      vals.extend([(str(mol["molecule_structures"][tag]) if "molecule_structures" in mol and mol["molecule_structures"] and tag in mol["molecule_structures"] else "") for tag in struct_tags])
      vals.extend([(str(mol["molecule_properties"][tag]) if "molecule_properties" in mol and mol["molecule_properties"] and tag in mol["molecule_properties"] else "") for tag in prop_tags])
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
      if nmax and n_mol>=nmax: break
    if nmax and n_mol>=nmax: break
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="mols")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}")

#############################################################################
def ListDrugs(skip, nmax, api_host=API_HOST, api_base_path=API_BASE_PATH, fout=None):
  n_mol=0; n_out=0; n_err=0; tags=None; struct_tags=None; prop_tags=None; tq=None;
  url_next = (f"{api_base_path}/drug.json?limit={NCHUNK}&offset={skip}")
  while True:
    rval = rest.GetURL("https://"+api_host+url_next, parse_json=True)
    if not rval: break
    mols = rval["drugs"] if "drugs" in rval else []
    for mol in mols:
      n_mol+=1
      logging.debug(json.dumps(mol, sort_keys=True, indent=2))
      if not tags:
        tags = sorted(mol.keys())
        for tag in tags[:]:
          if type(mol[tag]) in (dict, list, tuple):
            tags.remove(tag)
            logging.debug(f'Ignoring field ({type(mol[tag])}): "{tag}"')
        struct_tags = sorted(mol["molecule_structures"].keys())
        struct_tags.remove("molfile")
        prop_tags = sorted(mol["molecule_properties"].keys())
        fout.write('\t'.join(tags+struct_tags+prop_tags+["parent_chembl_id"])+'\n')
      vals = [(mol["molecule_hierarchy"]["parent_chembl_id"] if "molecule_hierarchy" in mol and mol["molecule_hierarchy"] and "parent_chembl_id" in mol["molecule_hierarchy"] else "")]
      vals.extend([(str(mol[tag]) if tag in mol else "") for tag in tags])
      vals.extend([(str(mol["molecule_structures"][tag]) if "molecule_structures" in mol and mol["molecule_structures"] and tag in mol["molecule_structures"] else "") for tag in struct_tags])
      vals.extend([(str(mol["molecule_properties"][tag]) if "molecule_properties" in mol and mol["molecule_properties"] and tag in mol["molecule_properties"] else "") for tag in prop_tags])
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
      if tq is not None: tq.update()
      if nmax and n_mol>=nmax: break
    if nmax and n_mol>=nmax: break
    total_count = rval["page_meta"]["total_count"] if "page_meta" in rval and "total_count" in rval["page_meta"] else None
    if not tq: tq = tqdm.tqdm(total=total_count, unit="drugs")
    url_next = rval["page_meta"]["next"] if "page_meta" in rval and "next" in rval["page_meta"] else None
    if not url_next: break
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}")

##############################################################################
def SearchMoleculeByName(ids, base_url=BASE_URL, fout=None):
  """IDs should be names/synonyms."""
  n_out=0; n_notfound=0; synonym_tags=None;
  tags = ["molecule_chembl_id"]
  for id_this in ids:
    rval = rest.GetURL(f"{base_url}/molecule/search?q={urllib.parse.quote(id_this)}", headers={"Accept":"application/json"}, parse_json=True)
    if not rval:
      logging.info(f'Not found: "{id_this}"')
      n_notfound+=1
      continue
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    mols = rval["molecules"] if "molecules" in rval else []
    for mol in mols:
      logging.debug(json.dumps(mol, sort_keys=True, indent=2))
      synonyms = mol["molecule_synonyms"] if "molecule_synonyms" in mol else []
      for synonym in synonyms:
        if not synonym_tags:
          synonym_tags = list(synonym.keys())
          fout.write('\t'.join(tags+synonym_tags)+'\n')

        molecule_synonym = synonym["molecule_synonym"] if "molecule_synonym" in synonym else ""
        if not re.search(id_this, molecule_synonym, re.I):
          continue

        vals = [(mol["molecule_chembl_id"] if "molecule_chembl_id" in mol else "")]
        vals.extend([(str(synonym[tag]) if tag in synonym else "") for tag in synonym_tags])
        fout.write(('\t'.join(vals))+'\n')
        n_out+=1
  logging.info(f"n_in: {len(ids)}; n_found: {len(ids)-n_notfound}; n_out: {n_out}")

#############################################################################
def GetMoleculeByInchikey(ids, base_url=BASE_URL, fout=None):
  """Requires InChI key, e.g. "GHBOEFUAGSHXPO-XZOTUCIWSA-N"."""
  n_out=0; tags=[]; struct_tags=[]; df=None; tq=None;
  for id_this in ids:
    if not tq: tq = tqdm.tqdm(total=len(ids), unit="mols")
    tq.update()
    mol = rest.GetURL(f"{base_url}/molecule/{id_this}.json", parse_json=True)
    if not mol:
      continue
    struct = mol["molecule_structures"] if "molecule_structures" in mol else None
    if not struct: continue
    if not tags:
      for tag in mol.keys():
        if type(mol[tag]) not in (list,dict): tags.append(tag)
      for tag in struct.keys():
        if type(struct[tag]) not in (list,dict): struct_tags.append(tag)
      struct_tags.remove("molfile")
    df_this = pd.concat([
	pd.DataFrame({tag:[(mol[tag] if tag in mol else None)] for tag in tags}),
	pd.DataFrame({tag:[(struct[tag] if tag in struct else None)] for tag in struct_tags})],
	axis=1)
    if fout is None:
      df = pd.concat([df, df_this])
    else:
      df_this.to_csv(fout, "\t", index=False)
    n_out+=1
  if tq is not None: tq.close()
  logging.info(f"n_qry: {len(ids)}; n_out: {n_out}; n_not_found: {len(ids)-n_out}")
  if fout is None: return df

#############################################################################



#!/usr/bin/env python3
"""
Pubtator REST API client
https://www.ncbi.nlm.nih.gov/CBBresearch/Lu/Demo/tmTools/RESTfulAPIs.html
Formats: JSON, PubTator, BioC.

Nomenclatures:
  Gene : NCBI Gene
e.g. https://www.ncbi.nlm.nih.gov/sites/entrez?db=gene&term=145226
  Disease : MEDIC (CTD, CTD_diseases.csv)
e.g. http://ctdbase.org/basicQuery.go?bqCat=disease&bq=C537775
  Chemical : MESH
e.g. http://www.nlm.nih.gov/cgi/mesh/2014/MB_cgi?field=uid&term=D000596
  Species : NCBI Taxonomy
e.g. https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?name=10090
  Mutation : tmVar
https://www.ncbi.nlm.nih.gov/CBBresearch/Lu/Demo/PubTator/tutorial/tmVar.html

NOTE that the API does NOT provide keyword search capability like
webapp https://www.ncbi.nlm.nih.gov/CBBresearch/Lu/Demo/PubTator/index.cgi
"""
import sys,os,time,json,argparse,re,logging
#
from .. import pubtator
#
API_HOST="www.ncbi.nlm.nih.gov"
API_BASE_PATH="/CBBresearch/Lu/Demo/RESTful/tmTool.cgi"
#
#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='PubTator REST API client', epilog='Reports PubMed NER annotations for specified PMID[s].')
  ops=['get_annotations']
  modes = ['Gene', 'Chemical', 'BioConcept']
  parser.add_argument("op", choices=ops, help="operation")
  parser.add_argument("--mode", choices=modes, help='mode', default='BioConcept')
  parser.add_argument("--ids", help="PubMed IDs, comma-separated (ex:25533513)")
  parser.add_argument("--i", dest="ifile", help="input file, PubMed IDs")
  parser.add_argument("--nmax", help="list: max to return")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  ids=[];
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    logging.info('Input IDs: %d'%(len(ids)))
    fin.close()
  elif args.ids:
    ids = re.split(r'[\s,]+', args.ids.strip())

  if args.op == 'get_annotations':
    if not ids: logging.error('Input PMIDs required.')
    pubtator.GetAnnotations(BASE_URL, args.mode, ids, fout)

  else:
    logging.error('Invalid operation: {0}'.format(args.op))



#!/usr/bin/env python3
"""
UMLS REST API client
UTS = UMLS Technology Services

 https://utslogin.nlm.nih.gov
 https://github.com/HHS/uts-rest-api
 https://documentation.uts.nlm.nih.gov/rest/home.html
 https://documentation.uts.nlm.nih.gov/rest/authentication.html
 https://documentation.uts.nlm.nih.gov/rest/concept/
 https://documentation.uts.nlm.nih.gov/rest/source-asserted-identifiers/
 https://documentation.uts.nlm.nih.gov/rest/search/
 https://www.nlm.nih.gov/research/umls/knowledge_sources/metathesaurus/release/abbreviations.html

 TGT = Ticket Granting Ticket
 (API requires one ticket per request.)
 CUI = Concept Unique Identifier
 Atom is a term in a source.
 Term-to-concept is many to one.

 Retrieves information for a known Semantic Type identifier (TUI)
 /semantic-network/{version}/TUI/{id}
 (DOES NOT SEARCH FOR INSTANCES OF THIS TYPE -- RETURNS METADATA ONLY.)
 Example TUIs:
 CHEM|Chemicals & Drugs|T103|Chemical
 CHEM|Chemicals & Drugs|T200|Clinical Drug
 CHEM|Chemicals & Drugs|T126|Enzyme
 CHEM|Chemicals & Drugs|T125|Hormone
 CHEM|Chemicals & Drugs|T129|Immunologic Factor
 CHEM|Chemicals & Drugs|T127|Vitamin
 DISO|Disorders|T020|Acquired Abnormality
 DISO|Disorders|T190|Anatomical Abnormality
 DISO|Disorders|T049|Cell or Molecular Dysfunction
 DISO|Disorders|T047|Disease or Syndrome
 DISO|Disorders|T048|Mental or Behavioral Dysfunction
 DISO|Disorders|T046|Pathologic Function
 DISO|Disorders|T184|Sign or Symptom
 GENE|Genes & Molecular Sequences|T087|Amino Acid Sequence
 GENE|Genes & Molecular Sequences|T088|Carbohydrate Sequence
 GENE|Genes & Molecular Sequences|T028|Gene or Genome
 GENE|Genes & Molecular Sequences|T085|Molecular Sequence
 GENE|Genes & Molecular Sequences|T086|Nucleotide Sequence
"""
###
import sys,os,argparse,re,yaml,json,csv,logging,requests,time
#
from .. import umls
#
#############################################################################
if __name__=='__main__':
  SRCS_PREFERRED = "ATC,HPO,ICD10,ICD10CM,ICD9CM,MDR,MSH,MTH,NCI,OMIM,RXNORM,SNOMEDCT_US,WHO"
  EPILOG = f"""\
All get* operations require --idsrc CUI, CUIs as inputs; CUI = Concept Unique Identifier.
Example CUIs: C34488, C0018787, C0016644
"""
  parser = argparse.ArgumentParser(description='UMLS REST API client utility', epilog=EPILOG)
  ops = ['getCodes', 'getAtoms', 'getRelations', 'listSources', 'xrefConcept', 'search', 'searchByTUI']
  searchTypes = ['exact', 'words', 'leftTruncation', 'rightTruncation', 'approximate', 'normalizedString']
  inputTypes = ['atom', 'code', 'sourceConcept', 'sourceDescriptor', 'sourceUi', 'tty']
  returnIdTypes = ['aui', 'concept', 'code', 'sourceConcept', 'sourceDescriptor', 'sourceUi']
  parser.add_argument("op", choices=ops, help='OPERATION (select one)')
  parser.add_argument("--ids", help="IDs (comma-separated) (ex:C0018787)")
  parser.add_argument("--idfile", help="input IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--idsrc", default="CUI", help="query ID source (default: CUI)")
  parser.add_argument("--searchType", choices=searchTypes, default='words', help=f" [words]")
  parser.add_argument("--inputType", choices=inputTypes, default='atom', help=f" [atom]")
  parser.add_argument("--returnIdType", choices=returnIdTypes, default='concept', help=f" [concept]")
  parser.add_argument("--srcs", default=SRCS_PREFERRED, help=f"sources to include in response [{SRCS_PREFERRED}]")
  parser.add_argument("--searchQuery", help='string or code')
  parser.add_argument("--skip", default=0, type=int)
  parser.add_argument("--nmax", default=0, type=int)
  parser.add_argument("--api_version", default=f"{umls.API_VERSION}", help=f"API version {umls.API_VERSION}")
  parser.add_argument("--api_host", default=umls.API_HOST)
  parser.add_argument("--api_base_path", default=umls.API_BASE_PATH)
  parser.add_argument("--api_auth_host", default=umls.API_AUTH_HOST)
  parser.add_argument("--api_auth_endpoint", default=umls.API_AUTH_ENDPOINT)
  parser.add_argument("--api_auth_service", default=umls.API_AUTH_SERVICE)
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.umls.yaml")
  parser.add_argument("--api_key", help="API key")
  parser.add_argument("-v", "--verbose", default=0, action="count")

  args = parser.parse_args()

  if args.op[:3] == 'get' and args.idsrc!='CUI':
    parser.error('Operation "{}" requires --idsrc CUI.'.format(args.op))

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  params = umls.ReadParamFile(args.param_file)
  if args.api_key: params['API_KEY'] = args.api_key
  if not params['API_KEY']:
    parser.error('Please specify valid API_KEY via --api_key or --param_file') 

  api_auth_url = 'https://'+args.api_auth_host+args.api_auth_endpoint
  auth = umls.Authentication(params['API_KEY'],
args.api_auth_service, api_auth_url, umls.API_HEADERS)
  auth.setVerbosity(args.verbose)

  ids=[];
  if args.idfile:
    fin = open(args.idfile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]', args.ids)
  if ids: logging.info(f'input IDs: {len(ids)}')

  t0 = time.time()

  srclist = umls.SourceList()
  srclist.initFromApi(base_url, args.api_version, auth)

  if args.srcs:
    for src in re.split(r'[,\s]+', args.srcs.strip()):
      if not srclist.has_src(src):
          parser.error(f'Source unknown: "{src}"')

  if args.op == 'listSources':
    fout.write("abbreviation\tshortName\tpreferredName\n")
    for i,src in enumerate(srclist.sources):
      abbr,name,prefname = src
      fout.write(f'{abbr}\t{name}\t{prefname}\n')
    logging.info(f'n_src: {len(srclist.sources)}')

  elif args.op == 'xrefConcept':
    umls.XrefConcept(args.idsrc, ids, args.skip, args.nmax, auth, args.api_version, base_url, fout)

  elif args.op == 'getRelations':
    umls.GetRelations(ids, args.skip, args.nmax, args.srcs, auth, args.api_version, base_url, fout)

  elif args.op == 'getAtoms':
    umls.GetAtoms(ids, args.skip, args.nmax, args.srcs, auth, args.api_version, base_url, fout)

  elif args.op == 'getCodes':
    umls.GetCodes(ids, args.srcs, auth, args.api_version, base_url, fout)

  elif args.op == 'search':
    if not args.searchQuery:
      parser.error('search requires --searchQuery')
    umls.Search(args.searchQuery, args.searchType, args.inputType, args.returnIdType, args.srcs, auth, args.api_version, base_url, fout)

  elif args.op == 'searchByTUI':
    parser.error(f'{args.op} NOT IMPLEMENTED YET.')

  else:
    parser.error(f'Invalid operation: {args.op}')

  logging.info(f"""Elapsed time: {time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))}""")



#!/usr/bin/env python3
'''
Parse, process Entrez PubMed XML (summaries or full), normally obtained via Entrez eUtils,
e.g. entrez_pubmed_search.pl.

Note that other Entrez XML (e.g. PubChem) very similar.
'''
import os,sys,re,time,argparse,logging

from xml.etree import ElementTree
from xml.parsers import expat

from .. import util

#############################################################################
def PubMed_Docsum2Tsv(tree, fout):
  '''From summary xml, extract 1st occurance of selected fields.'''
  docsums = util.xml.XpathFind('/eSummaryResult/DocSum', tree)
  tags = ['PubDate', 'EPubDate', 'Source', 'Title', 'ISSN', 'PubTypeList', 'FullJournalName',
	'LastAuthor', 'SO', 'DOI', 'ELocationID',
	'pmc'] ##pmc is special
  fout.write(('\t'.join(['Id']+tags))+'\n')
  for docsum in docsums:
    pubdata = {}
    for cnode in docsum.childNodes:
      if cnode.nodeName == 'Id':
        pubdata['Id'] = util.xml.DOM_NodeText(cnode)
      for tag in tags:
        if cnode.nodeName == 'Item' and util.xml.DOM_GetNodeAttr(cnode, 'Name')==tag:
          pubdata[tag] = util.xml.DOM_NodeText(cnode)
        if cnode.nodeName == 'Item' and util.xml.DOM_GetNodeAttr(cnode, 'Name')=="ArticleIds":
          for cnode2 in cnode.childNodes:
            if cnode2.nodeName == 'Item' and util.xml.DOM_GetNodeAttr(cnode2, 'Name')=="pmc":
              pubdata['pmc'] = util.xml.DOM_NodeText(cnode2)
    for j,tag in enumerate(['Id']+tags):
      fout.write('\t' if j>0 else '')
      fout.write('"%s"'%(pubdata[tag] if pubdata.has_key(tag) else ''))
    fout.write('\n')
  logging.info('DocSum count: %d'%len(docsums))
  return

#############################################################################
def PubMed_Doc2Tsv(tree, fout):
  '''From full xml, extract 1st occurance of selected fields.'''
  citations = util.xml.XpathFind('/PubmedArticleSet/PubmedArticle/MedlineCitation', tree)
  xps = [	#relative xpaths
	'PMID',
	'Article/Journal/JournalIssue/PubDate/Year',
	'Article/Journal/Title',
	'Article/Journal/ISSN',
	'Article/PublicationTypeList/PublicationType',
	'Article/ArticleTitle',
	#'Article/Abstract/AbstractText',
	'Article/AuthorList/Author/LastName'
	]
  fout.write(('\t'.join(map(lambda p:os.path.basename(p), xps)))+'\n')
  for citation in citations:
    for j,xp in enumerate(xps):
      nodes = util.xml.XpathFind(xp, citation)
      datum = util.xml.DOM_NodeText(nodes[0]) if nodes else ''
      fout.write('\t' if j>0 else '')
      fout.write('"%s"'%(datum))
    fout.write('\n')
  logging.info('citation count: %d'%len(citations))
  return

#############################################################################
def PubMed_Doc2AuthorList(tree, fout):
  '''From full xml for one pub, extract author list.'''
  authors = util.xml.XpathFind('/PubmedArticleSet/PubmedArticle/MedlineCitation/Article/AuthorList/Author', tree)
  for author in authors:
    n = util.xml.XpathFind('LastName', author)
    lname = util.xml.DOM_NodeText(n[0])
    fname = util.xml.DOM_NodeText(util.xml.XpathFind('ForeName', author)[0])
    affiliations = util.xml.XpathFind('AffiliationInfo/Affiliation', author)
    affs=[]
    for affiliation in affiliations:
      affs.append(util.xml.DOM_NodeText(affiliation))
    fout.write('"%s"\t"%s"\t"%s"\n'%(lname, fname, ('; '.join(affs))))
  logging.info('n_authors: %d'%len(authors))

#############################################################################
def PubMed_Docsum2AbstractText(tree, pmid, fout, verbose):
  '''From summary xml, extract title and abstract.'''
  articles = util.xml.XpathFind('/PubmedArticleSet/PubmedArticle', tree)
  title=None; abstract=None;
  for article in articles:
    try:
      pmid_this = int(util.xml.DOM_GetLeafValsByTagName(article, 'PMID')[0])
    except Exception as e:
      logging.error('PMID not found: %s'%str(e))
      continue
    if pmid_this==int(pmid):
      try:
        title = util.xml.DOM_GetLeafValsByTagName(article, 'ArticleTitle')[0]
      except Exception as e:
        logging.info('ERROR: ArticleTitle not found: %s'%str(e))
      try:
        abstract = util.xml.DOM_GetLeafValsByTagName(article, 'AbstractText')[0]
      except Exception as e:
        logging.info('ERROR: AbstractText not found: %s'%str(e))
      break
  if title or abstract:
    if not title: logging.info('ERROR: no title found for PMID: %d'%(pmid))
    if not abstract: logging.info('ERROR: no abstract found for PMID: %d ; title: "%s"'%(pmid, title))
    fout.write('%s\n%s\n'%(title, abstract))
    return True
  else:
    return False

#############################################################################
def PubMed_Doc2AbstractText(tree, pmid, fout, verbose):
  '''From full xml, extract title and abstract.'''
  citations = util.xml.XpathFind('/PubmedArticleSet/PubmedArticle/MedlineCitation', tree)
  title=None; abstract=None;
  for citation in citations:
    try:
      pmid_this = int(util.xml.DOM_GetLeafValsByTagName(citation, 'PMID')[0])
    except Exception as e:
      logging.info('ERROR: PMID not found: %s'%str(e))
      continue
    if pmid_this==int(pmid):
      try:
        title = util.xml.DOM_GetLeafValsByTagName(citation, 'ArticleTitle')[0]
      except Exception as e:
        logging.info('ERROR: ArticleTitle not found: %s'%str(e))
      try:
        abstract = util.xml.DOM_GetLeafValsByTagName(citation, 'AbstractText')[0]
      except Exception as e:
        logging.info('ERROR: AbstractText not found: %s'%str(e))
      break
  if title or abstract:
    if not title: logging.info('ERROR: no title found for PMID: %d'%(pmid))
    if not abstract: logging.info('ERROR: no abstract found for PMID: %d ; title: "%s"'%(pmid, title))
    fout.write('%s\n%s\n'%(title, abstract))
    return True
  else:
    return False

#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='process PubMed XML (summaries or full), typically obtained via Entrez eUtils.')
  ops=['summary2tsv', 'summary2abstract', 'full2tsv', 'full2abstract', 'full2authorlist']
  parser.add_argument("op", choices=ops, help="operation")
  parser.add_argument("--i", dest="ifile", required=True, help="input file, XML")
  parser.add_argument("--ids", help="PubMed IDs, comma-separated (ex:25533513)")
  parser.add_argument("--idfile", help="input file, PubMed IDs")
  parser.add_argument("--nmax", help="max to return")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--odir", help="output directory")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fin = open(args.ifile)

  pmids=[]
  if args.idfile:
    fin_ids = open(args.idfile)
    if not fin_ids: parser.error('cannot open idfile: %s'%args.idfile)
    while True:
      line = fin_ids.readline()
      if not line: break
      try:
        ids.append(int(line.rstrip()))
      except:
        logging.error('bad input ID: %s'%line)
        continue
    if args.verbose:
      logging.info('input IDs: %d'%(len(ids)))
    fin_ids.close()

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  try:
    tree = ElementTree.parse(fin)
  except Exception as e:
    parser.error('failed to parse XML: %s'%str(e))

  if args.op == 'summary2tsv':
    PubMed_Docsum2Tsv(tree, fout)

  elif args.op == 'full2tsv':
    PubMed_Doc2Tsv(tree, fout)

  elif args.op == 'full2authorlist':
    PubMed_Doc2AuthorList(tree, fout)

  elif args.op == 'summary2abstract':
    if not ids: parser.error('--summary2abstract requires --ids.')
    n_out=0;
    for id_this in ids:
      logging.info('%08d'%(id_this))
      if args.odir: fout = open('%s/%08d_title_and_abstract.txt'%(args.odir, id_this), "w")
      ok = PubMed_Docsum2AbstractText(tree, id_this, fout, args.verbose)
      if ok: n_out+=1
      else: logging.error('no title or abstract found for PMID: %d'%(id_this))
      if odir: fout.close()
    logging.info('abstracts found: %d'%n_out)

  elif args.op == 'full2abstract':
    if not ids: parser.error('--full2abstract requires --ids.')
    n_out=0;
    for id_this in ids:
      if args.verbose:
        logging.info('%08d'%(id_this))
      if args.odir: fout = open('%s/%08d_title_and_abstract.txt'%(odir, id_this), "w")
      ok = PubMed_Doc2AbstractText(tree, id_this, fout, args.verbose)
      if ok: n_out+=1
      else: logging.error('no title or abstract found for PMID: %d'%(id_this))
      if args.odir: fout.close()
    logging.info('abstracts found: %d'%n_out)

  else:
    parser.error('no operation specified.')



#!/usr/bin/env python3
"""
TIN-X (Target Importance and Novelty Explorer) REST API client
https://www.newdrugtargets.org/
https://api.newdrugtargets.org/docs
https://api.newdrugtargets.org/targets/
"""
###
import sys,os,re,json,time,logging,tqdm
import pandas as pd
import urllib,urllib.parse
#
from ...util import rest
#
API_HOST="api.newdrugtargets.org"
API_BASE_PATH=""
BASE_URL = 'https://'+API_HOST+API_BASE_PATH
#
NCHUNK=100;
#
##############################################################################
def ListTargets(base_url=BASE_URL, fout=None):
  n_out=0; tags=None; tq=None; df=None; offset=0;
  while True:
    url_next = (base_url+f'/targets/?limit={NCHUNK}&offset={offset}')
    rval = rest.Utils.GetURL(url_next, parse_json=True)
    targets = rval["results"] if "results" in rval else []
    if not tq: tq = tqdm.tqdm(total=rval["count"], unit="targets")
    for target in targets:
      tq.update()
      logging.debug(json.dumps(target, sort_keys=True, indent=2))
      if not tags: tags = list(target.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[target[tags[j]]] for j in range(len(tags))})])
      n_out+=1
    url_next = rval["next"] if "next" in rval else None #Full URL suboptimal.
    if not url_next: break
    offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)

##############################################################################
def SearchTargets(query_term, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  url_next = (base_url+'/targets/?search='+urllib.parse.quote(query_term))
  while True:
    rval = rest.Utils.GetURL(url_next, parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    targets = rval["results"] if "results" in rval else []
    for target in targets:
      logging.debug(json.dumps(target, sort_keys=True, indent=2))
      if not tags: tags = list(target.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[target[tags[j]]] for j in range(len(tags))})])
      n_out+=1
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)

##############################################################################
def ListDiseases(base_url=BASE_URL, fout=None):
  n_out=0; tags=None; tq=None; df=None; offset=0;
  while True:
    url_next = (base_url+f'/diseases/?limit={NCHUNK}&offset={offset}')
    rval = rest.Utils.GetURL(url_next, parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    diseases = rval["results"] if "results" in rval else []
    if not tq: tq = tqdm.tqdm(total=rval["count"], unit="dtos")
    for disease in diseases:
      tq.update()
      logging.debug(json.dumps(disease, sort_keys=True, indent=2))
      if not tags: tags = list(disease.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[disease[tags[j]]] for j in range(len(tags))})])
      n_out+=1
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
    offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)

##############################################################################
def ListArticles(base_url=BASE_URL, fout=None):
  n_out=0; tags=None; tq=None; df=None; offset=0;
  while True:
    url_next = (base_url+f'/articles/?limit={NCHUNK}&offset={offset}')
    rval = rest.Utils.GetURL(url_next, parse_json=True)
    articles = rval["results"] if "results" in rval else []
    if not tq: tq = tqdm.tqdm(total=rval["count"], unit="articles")
    for article in articles:
      tq.update()
      logging.debug(json.dumps(article, sort_keys=True, indent=2))
      if not tags: tags = list(article.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[article[tags[j]]] for j in range(len(tags))})])
      n_out+=1
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
    offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)

##############################################################################
def ListDTO(base_url=BASE_URL, fout=None):
  n_out=0; tags=None; tq=None; df=None; offset=0;
  while True:
    url_next = (base_url+f'/dto/?limit={NCHUNK}&offset={offset}')
    rval = rest.Utils.GetURL(url_next, parse_json=True)
    dtos = rval["results"] if "results" in rval else []
    if not tq: tq = tqdm.tqdm(total=rval["count"], unit="dtos")
    for dto in dtos:
      tq.update()
      logging.debug(json.dumps(dto, sort_keys=True, indent=2))
      if not tags: tags = list(dto.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[dto[tags[j]]] for j in range(len(tags))})])
      n_out+=1
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
    offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)

##############################################################################
def GetDisease(ids, base_url=BASE_URL, fout=None):
  """IDs should be TIN-X disease IDs, e.g. 5391."""
  n_in=0; n_out=0; tags=None; df=None;
  for id_this in ids:
    n_in+=1
    disease = rest.Utils.GetURL(base_url+f'/diseases/{id_this}/', parse_json=True)
    logging.debug(json.dumps(disease, sort_keys=True, indent=2))
    if not tags: tags = list(disease.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[disease[tags[j]]] for j in range(len(tags))})])
    n_out+=1
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_in: {n_in}; n_out: {n_out}")
  return(df)

##############################################################################
def GetDiseaseByDOId(ids, base_url=BASE_URL, fout=None):
  """IDs should be Disease Ontology IDs, e.g. DOID:9297."""
  n_in=0; n_out=0; tags=None; df=None;
  for id_this in ids:
    n_in+=1
    rval = rest.Utils.GetURL(base_url+f'/diseases/?doid={id_this}', parse_json=True)
    diseases = rval["results"] if "results" in rval else []
    for disease in diseases:
      logging.debug(json.dumps(disease, sort_keys=True, indent=2))
      if not tags: tags = list(disease.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[disease[tags[j]]] for j in range(len(tags))})])
      n_out+=1
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_in: {n_in}; n_out: {n_out}")
  return(df)

##############################################################################
def GetTarget(ids, base_url=BASE_URL, fout=None):
  """IDs should be TIN-X target IDs, e.g. 10027."""
  n_in=0; n_out=0; tags=None; df=None;
  for id_this in ids:
    n_in+=1
    target = rest.Utils.GetURL(base_url+f'/targets/{id_this}/', parse_json=True)
    logging.debug(json.dumps(target, sort_keys=True, indent=2))
    if not tags: tags = list(target.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[target[tags[j]]] for j in range(len(tags))})])
    n_out+=1
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_in: {}; n_out: {}".format(n_in, n_out))
  return(df)

##############################################################################
def GetTargetByUniprot(ids, base_url=BASE_URL, fout=None):
  """IDs should be UniProt IDs, e.g. Q9H4B4."""
  n_in=0; n_out=0; tags=None; df=None;
  for id_this in ids:
    n_in+=1
    rval = rest.Utils.GetURL(base_url+f'/targets/?uniprot={id_this}', parse_json=True)
    targets = rval["results"] if "results" in rval else []
    for target in targets:
      logging.debug(json.dumps(target, sort_keys=True, indent=2))
      if not tags: tags = list(target.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[target[tags[j]]] for j in range(len(tags))})])
      n_out+=1
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_in: {}; n_out: {}".format(n_in, n_out))
  return(df)

##############################################################################
def GetTargetDiseases(ids, base_url=BASE_URL, fout=None):
  """IDs should be TIN-X target IDs, e.g. 10027."""
  n_in=0; n_out=0; tags=None; tq=None; df=None;
  for id_this in ids:
    n_in+=1
    offset=0;
    while True:
      url_next = (base_url+f'/targets/{id_this}/diseases?limit={NCHUNK}&offset={offset}')
      rval = rest.Utils.GetURL(url_next, parse_json=True)
      diseases = rval["results"] if "results" in rval else []
      if not tq: tq = tqdm.tqdm(total=rval["count"], unit="diseases")
      for disease in diseases:
        tq.update()
        logging.debug(json.dumps(disease, sort_keys=True, indent=2))
        phenotype = disease["disease"] if "disease" in disease else {}
        if not tags:
          tags = list(disease.keys())
          tags.remove("disease")
          phenotype_tags = list(phenotype.keys())
        data_this = {"tinx_target_id":id_this}
        data_this.update({tags[j]:[disease[tags[j]]] for j in range(len(tags))})
        data_this.update({phenotype_tags[j]:[phenotype[phenotype_tags[j]]] for j in range(len(phenotype_tags))})
        df = pd.concat([df, pd.DataFrame(data_this)])
        n_out+=1
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
      if n_out>=rval["count"]: break #url_next may be wrong.
      offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_in: {n_in}; n_out: {n_out}")
  return(df)

##############################################################################
def GetDiseaseTargets(ids, base_url=BASE_URL, fout=None):
  """IDs should be TIN-X disease IDs, e.g. 5391."""
  n_in=0; n_out=0; tags=None; tq=None; df=None;
  for id_this in ids:
    n_in+=1
    offset=0;
    while True:
      url_next = (base_url+f'/diseases/{id_this}/targets?limit={NCHUNK}&offset={offset}')
      rval = rest.Utils.GetURL(url_next, parse_json=True)
      logging.debug(json.dumps(rval, sort_keys=True, indent=2))
      targets = rval["results"] if "results" in rval else []
      if not tq: tq = tqdm.tqdm(total=rval["count"], unit="targets")
      for target in targets:
        tq.update()
        logging.debug(json.dumps(target, sort_keys=True, indent=2))
        protein = target["target"] if "target" in target else {}
        if not tags:
          tags = list(target.keys())
          tags.remove("target")
          protein_tags = list(protein.keys())
        data_this = {"tinx_disease_id":id_this}
        data_this.update({tags[j]:[target[tags[j]]] for j in range(len(tags))})
        data_this.update({protein_tags[j]:[protein[protein_tags[j]]] for j in range(len(protein_tags))})
        df = pd.concat([df, pd.DataFrame(data_this)])
        n_out+=1
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
      if n_out>=rval["count"]: break #url_next may be wrong.
      offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_in: {n_in}; n_out: {n_out}")
  return(df)

##############################################################################
def GetDiseaseTargetArticles(disease_ids, ids, base_url=BASE_URL, fout=None):
  """IDs should be TIN-X disease and target IDs and, e.g. 5391, 12203."""
  n_out=0; tags=None; df=None;
  for did in disease_ids:
    for tid in ids:
      offset=0;
      while True:
        url_next = (base_url+'/diseases/{did}/targets/{tid}/articles?limit={NCHUNK}&offset={offset}')
        rval = rest.Utils.GetURL(url_next, parse_json=True)
        articles = rval["results"] if "results" in rval else []
        for article in articles:
          logging.debug(json.dumps(article, sort_keys=True, indent=2))
          if not tags: tags = list(article.keys())
          df = pd.concat([df, pd.DataFrame({tags[j]:[article[tags[j]]] for j in range(len(tags))})])
          n_out+=1
        url_next = rval["next"] if "next" in rval else None
        if not url_next: break
        if n_out>=rval["count"]: break #url_next may be wrong.
        offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)

##############################################################################
def SearchDiseases(query_term, base_url=BASE_URL, fout=None):
  """Search names; begins-with search logic."""
  n_out=0; tags=None; df=None; offset=0;
  while True:
    url_next = (base_url+'/diseases/?search={}&limit={}&offset={}'.format(urllib.parse.quote(query_term), NCHUNK, offset))
    rval = rest.Utils.GetURL(url_next, parse_json=True)
    diseases = rval["results"] if "results" in rval else []
    for disease in diseases:
      logging.debug(json.dumps(disease, sort_keys=True, indent=2))
      if not tags: tags = list(disease.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[disease[tags[j]]] for j in range(len(tags))})])
      n_out+=1
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
    offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)

##############################################################################
def SearchTargets(query_term, base_url=BASE_URL, fout=None):
  """Search names."""
  n_out=0; tags=None; df=None; offset=0;
  while True:
    url_next = (base_url+'/targets/?search={}&limit={}&offset={}'.format(urllib.parse.quote(query_term), NCHUNK, offset))
    rval = rest.Utils.GetURL(url_next, parse_json=True)
    targets = rval["results"] if "results" in rval else []
    for target in targets:
      logging.debug(json.dumps(target, sort_keys=True, indent=2))
      if not tags: tags = list(target.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[target[tags[j]]] for j in range(len(tags))})])
      n_out+=1
    url_next = rval["next"] if "next" in rval else None
    if not url_next: break
    offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)

##############################################################################
def SearchArticles(terms, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None; offset=0;
  for term in terms:
    while True:
      url_next = (base_url+'/articles/?search={}&limit={}&offset={}'.format(urllib.parse.quote(query_term), NCHUNK, offset))
      rval = rest.Utils.GetURL(url_next, parse_json=True)
      logging.debug(json.dumps(rval, sort_keys=True, indent=2))
      articles = rval["results"] if "results" in rval else []
      for article in articles:
        logging.debug(json.dumps(article, sort_keys=True, indent=2))
        if not tags: tags = list(article.keys())
        df = pd.concat([df, pd.DataFrame({tags[j]:[article[tags[j]]] for j in range(len(tags))})])
        n_out+=1
      url_next = rval["next"] if "next" in rval else None
      if not url_next: break
      offset += NCHUNK
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(n_out))
  return(df)




#!/usr/bin/env python3
###

import sys,os,time,logging
import numpy as np
import pandas as pd
import h5py

from ...util import hdf as hdf_util

#############################################################################
def ListSamples(f, fout):
  if "meta" not in f or "samples" not in f["meta"]:
    logging.error("No samples found.")
    return
  #samples_title = f["meta"]["samples/title"]
  #print(pd.DataFrame(samples_title.asstr()[0:10,]))
  samples = f["meta"]["samples"]
  df = pd.DataFrame()
  for i,k in enumerate(list(samples.keys())):
    logging.debug(f"{i+1}. {samples[k].name}")
    if type(samples[k]) is h5py.Dataset:
      df_this = pd.DataFrame(samples[k])
      df = pd.concat([df, df_this], axis=1)
      df = df.drop_duplicates()
    #if i>2: break #DEBUG
  df.columns = list(samples.keys())
  #print(df.iloc[0:10,:])

  for col, dtype in df.dtypes.items():
    if dtype == np.object:  # Only process object columns.
      # decode, or return original value if decode return Nan
      df[col] = df[col].str.decode('utf-8').fillna(df[col]) 

  logging.info(f"Output rows: {df.shape[0]}; columns: {df.shape[1]}")
  df.to_csv(fout, "\t", index=False)

#############################################################################



#!/usr/bin/env python3
#############################################################################
### CMap Clue API client.
### CMap is the project; Clue is the platform.
### https://clue.io/api
### https://clue.io/connectopedia/query_api_tutorial
### https://clue.io/connectopedia/perturbagen_types_and_controls
### l1000_type: "landmark"|"inferred"|"best inferred"|"not inferred"
### BING = Best inferred plus Landmark
### API services: cells, genes, perts, pcls, plates, profiles, sigs,
### probeset_to_entrez_id, rep_fda_exclusivity
#############################################################################
### Apps:
### https://clue.io/cell-app
### https://clue.io/repurposing-app
#############################################################################
### curl -X GET --header "user_key: YOUR_KEY_HERE" 'https://api.clue.io/api/cells?filter=\{"where":\{"provider_name":"ATCC"\}\}' |python3 -m json.tool
#############################################################################
import sys,os,re,argparse,json,logging
#
from ... import lincs
from ...util import yaml as util_yaml
#
#############################################################################
if __name__=="__main__":
  epilog='''\
CMap is the project; Clue is the platform.
https://clue.io/api. 
Credentials config file should be at $HOME/.clueapi.yaml.
'''
  parser = argparse.ArgumentParser(description='CLUE.IO REST API client utility', epilog=epilog)
  ops = ['getGenes', 'listGenes', 'listGenes_landmark', 
	'getPerturbagens', 'listPerturbagens', 'listDrugs',
	'countSignatures', 'getSignatures',
	'getCells', 'listCells',
	'listPerturbagenClasses',
	'listDatasets', 'listDatatypes',
	'search']
  id_types = ['cell_id', 'pert_id', 'gene_symbol', 'entrez_id']
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--ids", help="IDs, comma-separated")
  parser.add_argument("--i", dest="ifile", help="input file, IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--id_type", choices=id_types, help='query ID or field type, e.g. gene_symbol')
  parser.add_argument("--clue_where", help="Clue API search where, e.g. '{\"pert_desc\":\"sirolimus\",\"cell_id\":\"MCF7\"}'")
  parser.add_argument("--nmax", type=int, default=1000, help="max results")
  parser.add_argument("--skip", type=int, default=0, help="skip results")
  parser.add_argument("--api_host", default=lincs.clue.Utils.API_HOST)
  parser.add_argument("--api_base_path", default=lincs.clue.Utils.API_BASE_PATH)
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.clueapi.yaml")
  parser.add_argument("-v", "--verbose", dest="verbose", action="count", default=0)
  args = parser.parse_args()

  params = util_yaml.ReadParamFile(args.param_file)

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  if args.ifile:
    fin = open(args.ifile)
    ids=[];
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
    logging.info('input queries: %d'%(len(ids)))
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())

  if args.op=='getGenes':
    if not ids: parser.error('--ids or --i required.')
    lincs.clue.Utils.GetGenes(params, ids, args.id_type, base_url, fout)

  elif args.op=='listGenes':
    lincs.clue.Utils.ListGenes(params, base_url, fout)

  elif args.op=='listGenes_landmark':
    lincs.clue.Utils.ListGenes_Landmark(params, base_url, fout)

  elif args.op=='getPerturbagens':
    if not ids: parser.error('--ids or --i required.')
    lincs.clue.Utils.GetPerturbagens(params, ids, args.id_type, base_url, fout)

  elif args.op=='listPerturbagens':
    lincs.clue.Utils.ListPerturbagens(params, base_url, fout)

  elif args.op=='listDrugs':
    lincs.clue.Utils.ListDrugs(params, base_url, fout)

  elif args.op=='getCells':
    if not ids: parser.error('--ids or --i required.')
    lincs.clue.Utils.GetCells(params, ids, args.id_type, base_url, fout)

  elif args.op=='listCells':
    lincs.clue.Utils.ListCells(params, base_url, fout)

  elif args.op=='getSignatures':
    if not args.clue_where: parser.error('--clue_where required.')
    lincs.clue.Utils.GetSignatures(params, args, base_url, fout)

  elif args.op=='countSignatures':
    if not args.clue_where: parser.error('--clue_where required.')
    lincs.clue.Utils.CountSignatures(params, args, base_url, fout)

  elif args.op=='listDatasets':
    lincs.clue.Utils.ListDatasets(params, base_url, fout)

  elif args.op=='listDatatypes':
    lincs.clue.Utils.ListDatatypes(params, base_url, fout)

  elif args.op=='listPerturbagenClasses':
    lincs.clue.Utils.ListPerturbagenClasses(params, base_url, fout)

  else:
    parser.error('Unsupported operation: %s'%args.op)



#!/usr/bin/env python3
"""
Pandas CSV/TSV utilities.
"""
###
import sys,os,re,logging
import pandas as pd

#############################################################################
def SearchRows(df, cols, coltags, qrys, rels, typs, fout):
  n = df.shape[0]
  for j,tag in enumerate(df.columns):
    if cols:
      if j not in cols: continue
      else: jj = cols.index(j)
    elif coltags:
      if tag not in coltags: continue
      else: jj = coltags.index(tag)
    logging.debug(f"tag={tag}; j={j}; jj={jj}")
    if qrys[jj].upper() in ('NA','NAN'):
      df = df[df[tag].isna()]
    elif typs[jj]=='int':
      df = df[df[tag].astype('int')==int(qrys[jj])]
    elif typs[jj]=='float':
      df = df[df[tag].astype('float')==float(qrys[jj])]
    else:
      df = df[df[tag].astype('str').str.match('^'+qrys[jj]+'$')]
  df.to_csv(fout, '\t', index=False)
  logging.info(f"Rows found: {df.shape[0]} / {n}")

#############################################################################
def SampleRows(df, sample_frac, sample_n, fout):
  n = df.shape[0]
  if sample_n:
    df = df.sample(n=sample_n)
  else:
    df = df.sample(frac=sample_frac)
  df.to_csv(fout, '\t', index=False)
  logging.info(f"Rows sampled: {df.shape[0]} / {n}")

#############################################################################
def RemoveHeader(df, delim, fout):
  df.to_csv(fout, delim, index=False, header=False)

#############################################################################
def SetHeader(df, coltags, delim, fout):
  if not coltags:
    logging.error("Coltags required.")
  elif len(coltags)!=df.shape[1]:
    logging.error(f"len(coltags) != ncol ({len(coltags)} != {df.shape[1]}).")
  else:
    df.to_csv(fout, delim, index=False, header=coltags)

#############################################################################
def ToHtml(df, title, prettify, fout):
  precision=3;
  if prettify:
    from pandas.io.formats.style import Styler #Pandas v1.4.0+.
    styler = Styler(df, precision=precision)
    styler.set_caption(title if title is not None else (f"{df.shape[0]}rows,{df.shape[1]}cols"))
    styler.set_properties(**{'background-color': 'white', 'color':'navy', 'border':'1px solid black'})
    fout.write(styler.to_html())
  else:
    df.round(precision).to_html(fout, index=True, header=True,
	formatters=None, float_format=None, sparsify=None,
	bold_rows=True, border=None, justify=None,
	classes=None, render_links=True,
	show_dimensions=False)

#############################################################################



#!/usr/bin/env python3
"""
Functions for BRENDA API (SOAP) access.

"EC" refers to Enzyme Commission.  EC Numbers are assigned based on functionality,
the catalyzed reaction, not the protein itself.  Hence multiple proteins can 
exist for a given ECN or ECN+organism.

Default is human enzymes, organism="Homo sapiens".

Ref: http://www.brenda-enzymes.org/soap.php

Ref:	BRENDA: a resource for enzyme data and metabolic information.
	Schomburg I, Chang A, Hofmann O, Ebeling C, Ehrentreich F, Schomburg D.
	Trends Biochem Sci. 2002 Jan;27(1):54-6.
	http://www.ncbi.nlm.nih.gov/pubmed/11796225

Ref:	BRENDA in 2015: exciting developments in its 25th year of existence.
	Nucleic Acids Res. 2014 Nov 5. pii: gku1068. [Epub ahead of print]
	Chang A, Schomburg I, Placzek S, Jeske L, Ulbrich M, Xiao M, Sensen CW, Schomburg D.
	http://www.ncbi.nlm.nih.gov/pubmed/25378310.

Ref: http://en.wikipedia.org/wiki/Enzyme_inhibitor
"""
###
import sys,os,os.path,time,argparse,logging,hashlib
import string,re,json

from .. import brenda

#############################################################################
if __name__=='__main__':
  API_HOST = "www.brenda-enzymes.org"
  API_BASE_PATH = "/soap"
  API_WSDL_URL = "https://www.brenda-enzymes.org/soap/brenda_zeep.wsdl"

  epilog='''\
OPERATIONS:
    get = get record;
    get_names = get enzyme names;
    get_systematicname = get systematic enzyme name;
    get_organism = get organism;
    get_sequence = get sequence;
    get_pdb = get PDB code;
    get_ligands = get ligands code;
    get_inhibitors = get enzyme inhibitors;
    get_activators = get enzyme activating cpds;
    get_kivalues = get enzyme Ki values;
    get_kmvalues = get enzyme Km values;
    get_references = get references;
    get_inhibitordata = get inhibitor data (CSV);
    get_sequencedata = get sequence data (CSV);
    get_liganddata = get ligand data (CSV);
    get_referencedata = get reference data (CSV);
    list_all = list all ECNs (for specified organism);
    list_fromsynonyms = list ECNs linked to synonyms;
    list_frominhibitors = list ECNs linked to inhibitors;
    list_fromactivators = list ECNs linked to activating cpds;
    list_fromkivalues = list ECNs linked to Ki values;
    list_fromkmvalues = list ECNs linked to Km values;
    list_organisms = list all organisms;
Notes:
    "get*" operations (require --ecn or --ifile).;
    Km is concentration of substrate for half-max enzyme velocity, where velocity is rate of reaction product formation.;
    Ki is concentration of substrate for half-max enzyme inhibition.;
Example ECNs:
    1.9.3.1,
    1.14.99.1
    ;
Example Organisms:
    Homo sapiens,
    Mus musculus
    ;
'''
  IDTYPE='ecn';
  ORGANISM='Homo sapiens';

  parser = argparse.ArgumentParser(description='BRENDA API client', epilog=epilog)
  ops = ['test', 'get', 'get_names', 'get_systematicname', 'get_organism', 'get_sequence', 'get_pdb', 'get_ligands', 'get_inhibitors', 'get_activators', 'get_kivalues', 'get_kmvalues', 'get_references', 'get_inhibitordata', 'get_sequencedata', 'get_liganddata', 'get_referencedata', 'list_all', 'list_fromsynonyms', 'list_frominhibitors', 'list_fromactivators', 'list_fromkivalues', 'list_fromkmvalues', 'list_organisms']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="input file of ECNs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--ecns", help="ECNs (comma-separated)")
  parser.add_argument("--organism", default=ORGANISM, help="organism (may be 'all')")
  parser.add_argument("--nmax", type=int, default=None, help="max NMAX ECNs")
  parser.add_argument("--skip", type=int, default=0, help="skip 1st NSKIP ECNs")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("--api_wsdl_url", default=API_WSDL_URL)
  parser.add_argument("--api_user", help="API username")
  parser.add_argument("--api_key", help="API key")
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.brenda.yaml")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  params = drugcentral.ReadParamFile(args.param_file)
  if args.api_user: params['user_name'] = args.api_user
  if args.api_key: params['user_key'] = args.api_key

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  API_BASE_URL='http://'+args.api_host+args.api_base_path
  #API_BASE_URL='https://'+args.api_host+args.api_base_path

  CLIENT = brenda.SoapAPIClient(args.api_wsdl_url)
  if not CLIENT:
    logging.error("SOAP Client instantiation failed.")

  password = hashlib.sha256(params['user_key'].encode("utf-8")).hexdigest()
  API_PARAMS = [params['user_name'], password]

  organism = None if args.organism.lower()=='all' else args.organism

  fout = open(ofile, "w") if args.ofile else sys.stdout

  ecns=[]
  if args.ifile:
    fin = open(args.ifile)
    i=0;
    while True:
      line = fin.readline()
      if not line: break
      i+=1
      if skip>0 and i<skip: continue
      ecns.append(line.rstrip())
      if nmax>0 and len(ecns)==nmax: break
    logging.info('input ECNs: %d'%(len(ecns)))
    fin.close()
  elif args.ecns:
    ecns = re.split(r'[,\s]+', args.ecns)

  t0=time.time()

  if args.op[:4]=="get_" and not ecns: parser.error("Operation %s requires --ecns or --ifile."%args.op)

  if args.op=="get":
    brenda.GetECN(CLIENT, API_PARAMS, ecns, fout)

  elif args.op=="get_inhibitordata":
    brenda.GetInhibitorData(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_sequencedata":
    brenda.GetSequenceData(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_liganddata":
    brenda.GetLigandData(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_referencedata":
    brenda.GetReferenceData(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_names":
    brenda.GetNames(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_systematicname":
    brenda.GetSystematicName(CLIENT, API_PARAMS, ecns, fout)

  elif args.op=="get_organism":
    brenda.GetOrganism(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_sequence":
    brenda.GetSequence(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_pdb":
    brenda.GetPdb(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_kivalues":
    brenda.GetKiValues(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_kmvalues":
    brenda.GetKmValues(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_inhibitors":
    brenda.GetInhibitors(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_activators":
    brenda.GetActivators(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_ligands":
    brenda.GetLigands(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="get_references":
    brenda.GetReferences(CLIENT, API_PARAMS, ecns, args.organism, fout)

  elif args.op=="list_all":
    brenda.ListECNumbers(CLIENT, API_PARAMS, fout)

  elif args.op=="list_fromsynonyms":
    brenda.ListECNumbersFromSynonyms(CLIENT, API_PARAMS, fout)

  elif args.op=="list_frominhibitors":
    brenda.ListECNumbersFromInhibitors(CLIENT, API_PARAMS, fout)

  elif args.op=="list_fromactivators":
    brenda.ListECNumbersFromActivators(CLIENT, API_PARAMS, fout)

  elif args.op=="list_fromkivalues":
    brenda.ListECNumbersFromKiValue(CLIENT, API_PARAMS, fout)

  elif args.op=="list_fromkmvalues":
    brenda.ListECNumbersFromKmValue(CLIENT, API_PARAMS, fout)

  elif args.op=="list_organisms":
    brenda.ListOrganisms(CLIENT, API_PARAMS, fout)

  elif args.op=="test":
    brenda.Test(CLIENT, API_PARAMS)

  else:
    parser.error("Invalid operation: %s"%args.op)


  logging.info('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))))




#!/usr/bin/env python3
"""
 https://www.disgenet.org/api/
 https://www.disgenet.org/dbinfo
 DisGeNET Disease Types : disease, phenotype, group
 DisGeNET Metrics:
   GDA Score
   VDA Score
   Disease Specificity Index (DSI)
   Disease Pleiotropy Index (DPI)
   Evidence Level (EL)
   Evidence Index (EI)
 GNOMAD pLI (Loss-of-function Intolerant)

 DisGeNET Association Types:
   Therapeutic
   Biomarker
   Genomic Alterations
   GeneticVariation
   Causal Mutation
   Germline Causal Mutation
   Somatic Causal Mutation
   Chromosomal Rearrangement
   Fusion Gene
   Susceptibility Mutation
   Modifying Mutation
   Germline Modifying Mutation
   Somatic Modifying Mutation
   AlteredExpression
   Post-translational Modification
"""
###
import sys,os,re,argparse,time,logging
#
from .. import disgenet
from ..util import yaml as util_yaml
#
##############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(
        description='Disgenet (www.disgenet.org) REST API client utility',
        epilog="""\
Example IDs: C0019829 (UMLS CUI); 351 (NCBI Gene ID) """)
  ops = ['list_diseases', 'get_diseases', 'get_disease_GDAs', 'get_gene_GDAs', 'get_protein_GDAs']
  parser.add_argument("op",choices=ops,help='OPERATION')
  parser.add_argument("--ids", dest="ids", help="IDs, comma-separated (Disease UMLS CUIs, or Gene EntrezIDs, HGNC symbols, or UniProtIDs)")
  parser.add_argument("--i", dest="ifile", help="input file of IDs")
  parser.add_argument("--source", choices=disgenet.SOURCES, default="ALL", help="default=ALL")
  parser.add_argument("--disease_type", choices=disgenet.DISEASE_TYPES, default="disease")
  parser.add_argument("--disease_class", choices=disgenet.DISEASE_CLASSES, help="MeSH disease class")
  parser.add_argument("--o", dest="ofile", help="output file (TSV)")
  parser.add_argument("--nmax", type=int, default=None, help="max results")
  parser.add_argument("--api_host", default=disgenet.API_HOST)
  parser.add_argument("--api_base_path", default=disgenet.API_BASE_PATH)
  parser.add_argument("--user_email")
  parser.add_argument("--user_password")
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.disgenet.yaml")
  parser.add_argument("-v", "--verbose", action="count", default=0)

  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  if os.path.isfile(args.param_file):
    params = util_yaml.ReadParamFile(args.param_file)
  if args.user_email: params['EMAIL'] = args.user_email 
  if args.user_password: params['PASSWORD'] = args.user_password
  logging.debug(f"EMAIL: \"{params['EMAIL']}\"; PASSWORD: \"{params['PASSWORD']}\"")

  base_url='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile,"w") if args.ofile else sys.stdout

  t0=time.time()

  if args.ifile:
    fin = open(args.ifile)
    ids = [];
    while True:
      line = fin.readline()
      if not line: break
      if line.rstrip(): ids.append(line.rstrip())
    logging.info(f"Input queries: {len(ids)}")
    fin.close()
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())

  if args.op == "get_disease_GDAs":
    disgenet.GetDiseaseGDAs(ids, args.source, params["EMAIL"], params["PASSWORD"], base_url, fout)

  elif args.op == "get_gene_GDAs":
    disgenet.GetGeneGDAs(ids, args.source, params["EMAIL"], params["PASSWORD"], base_url, fout)

  elif args.op == "get_protein_GDAs":
    disgenet.GetProteinGDAs(ids, args.source, params["EMAIL"], params["PASSWORD"], base_url, fout)

  elif args.op == "get_diseases":
    disgenet.GetDiseases(args.source, args.disease_type, args.disease_class, args.nmax, params["EMAIL"], params["PASSWORD"], base_url, fout)

  elif args.op == "list_diseases":
    disgenet.ListDiseases(params["EMAIL"], params["PASSWORD"], base_url, fout)

  else:
    parser.error("No operation specified.")

  logging.info(f"elapsed time: {time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
Utility for ChEBI SOAP API.

* https://www.ebi.ac.uk/chebi/webServices.do
"""
###
import sys,os,re,json,argparse,time,logging
import pandas as pd
#
from .. import chebi
#
##############################################################################
if __name__=='__main__':
  epilog="Example entity IDs: 30273,33246,24433"
  parser = argparse.ArgumentParser(description='ChEBI SOAP API client', epilog=epilog)
  ops = [ 
	"get_entity",
	"get_entity_children",
	"get_entity_parents",
	"search"
	]
  parser.add_argument("op", choices=ops, help='OPERATION (select one)')
  parser.add_argument("--ids", help="input IDs")
  parser.add_argument("--i", dest="ifile", help="input file, IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--query", help="search query (SMILES)")
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--nmax", type=int, default=None)
  parser.add_argument("--api_host", default=chebi.API_HOST)
  parser.add_argument("--api_base_path", default=chebi.API_BASE_PATH)
  parser.add_argument("-v","--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, 'w') if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())
  if len(ids)>0: logging.info('Input IDs: %d'%(len(ids)))

  if args.op[:3]=="get" and not (args.ifile or args.ids):
    parser.error(f"--i or --ids required for operation {args.op}.")

  if args.op == "get_entity":
    chebi.GetEntity(ids, base_url, fout)

  elif args.op == "get_entity_children":
    chebi.GetEntityChildren(ids, base_url, fout)

  elif args.op == "get_entity_parents":
    chebi.GetEntityParents(ids, base_url, fout)

  elif args.op == "search":
    parser.error(f'Not yet implemented: {args.op}')
    #chebi.Search(args.query, base_url, fout)

  else:
    parser.error(f'Invalid operation: {args.op}')



#!/usr/bin/env python3
'''
Commonly used functions for REST client applications.

* JSON and XML handled, parsed into objects.
* HTTP headers and POST data handled.

'''
import sys,os,io,re,time,logging,base64,json
import urllib,urllib.request,urllib.parse
from xml.etree import ElementTree as ET
#
REST_TIMEOUT=10
REST_RETRY_NMAX=10
REST_RETRY_WAIT=5
#
##############################################################################
def GetURL(url, headers={}, parse_json=False, usr=None, pw=None, parse_xml=False, nmax_retry=REST_RETRY_NMAX, verbose=0):
  '''Entry point for GET requests.'''
  return RequestURL(url, headers, None, usr, pw, parse_json, parse_xml, nmax_retry, verbose)

##############################################################################
def PostURL(url, headers={}, data={}, usr=None, pw=None, parse_json=False, parse_xml=False, nmax_retry=REST_RETRY_NMAX, verbose=0):
  '''Entry point for POST requests.'''
  return RequestURL(url, headers, data, usr, pw, parse_json, parse_xml, nmax_retry, verbose)

##############################################################################
def RequestURL(url, headers, data, usr, pw, parse_json, parse_xml, nmax_retry, verbose):
  '''Use internally, not externally.  Called by GetURL() and PostURL().  Only Basic authentication supported.'''
  if data and type(data) is dict:
    data = urllib.parse.urlencode(data).encode('utf-8')
  elif data and type(data) in (str,): #Ok for SOAP/XML?
    data = data.encode('utf-8')
  else:
    data = None
  req = urllib.request.Request(url=url, headers=headers, data=data)
  if usr and pw:
    req.add_header("Authorization", "Basic %s"%base64.encodestring('%s:%s'%(usr,pw)).replace('\n','')) 

  logging.debug('url="%s"'%url)
  logging.debug('request type = %s'%req.type)
  logging.debug('request method = %s'%req.get_method())
  logging.debug('request host = %s'%req.host)
  logging.debug('request full_url = %s'%req.full_url)
  logging.debug('request header_items = %s'%req.header_items())
  if data:
    logging.debug('request data = %s'%req.data)
    logging.debug('request data size = %s'%len(req.data))

  i_try=0
  while True:
    i_try+=1
    try:
      with urllib.request.urlopen(req, timeout=REST_TIMEOUT) as f:
        fbytes = f.read() #With Python3 read bytes from sockets.
        ftxt = fbytes.decode('utf-8')
        #f.close()
    except urllib.request.HTTPError as e:
      if e.code==404:
        return ([])
      elif e.code==400:
        logging.warn('%s (URL=%s)'%(e, url))
      else:
        logging.warn('%s'%e)
      return None
    except urllib.request.URLError as e:  ## may be socket.error
      # may be "<urlopen error [Errno 110] Connection timed out>"
      logging.warn('[try %d/%d]: %s'%(i_try, nmax_retry, e))
      if i_try<nmax_retry:
        time.sleep(REST_RETRY_WAIT)
        continue
      return None
    except Exception as e:
      logging.warn('%s'%e)
      if i_try<nmax_retry:
        time.sleep(REST_RETRY_WAIT)
        continue
      return None
    break

  if ftxt.strip()=='': return None
  #logging.debug('%s'%ftxt)

  if parse_json:
    try:
      rval = json.loads(ftxt, encoding='utf-8')
    except ValueError as e:
      logging.debug('JSON Error: %s'%e)
      try:
        ### Should not be necessary.  Backslash escapes allowed in JSON.
        ftxt_fix = ftxt.replace(r'\"', '&quot;').replace(r'\\', '')
        ftxt_fix = ftxt_fix.replace(r'\n', '\\\\n') #ok?
        rval = json.loads(ftxt_fix, encoding='utf-8')
        logging.debug('Apparently fixed JSON Error: %s'%e)
        logging.debug('Apparently fixed JSON: "%s"'%ftxt)
      except ValueError as e:
        logging.error('Failed to fix JSON Error: %s'%e)
        logging.error('ftxt_fix="%s"'%ftxt_fix)
        raise
  elif parse_xml:
    try:
      rval = ParseXml(ftxt)
    except Exception as e:
      logging.error('XML Error: %s'%e)
      rval=ftxt
  else: #raw
    rval=ftxt

  return rval

#############################################################################
def ParseXml(xml_str):
  """ElementTree.fromstring() returns root Element. ElementTree.parse() returns ElementTree."""
  etree=None;
  try:
    etree = ET.parse(io.StringIO(xml_str))
  except Exception as e:
    logging.warn('XML parse error: %s'%e)
    return False
  return etree

#############################################################################



#!/usr/bin/env python
#############################################################################
### pubchem_assaysim.py - assay similarity based on activity profiles
### from PubChem CSV assay file[s]
#############################################################################
import os,sys,re,getopt,gzip,zipfile,tempfile
#try:
#  import gdbm
#except:
#  import dbm as gdbm

from ... import pubchem

PROG=os.path.basename(sys.argv[0])
DATADIR="/home/data/pubchem/bioassay/csv/data"

SCRATCHDIR='/tmp/'+PROG+'_SCRATCHDIR'

#############################################################################
if __name__=='__main__':
  usage='''
  %(PROG)s - assay similarity based on activity profiles (A vs. B)

  required:
  --aidA=<AID> ... assay ID A
  --aidB=<AID> ... assay ID B

  options:

  --v            ... verbose
  --vv           ... very verbose
  --h            ... this help
'''%{'PROG':PROG,'DATADIR':DATADIR}

  def ErrorExit(msg):
    print >>sys.stderr,msg
    sys.exit(1)

  indir=DATADIR;
  infile=None; verbose=0;
  aidA=None; aidB=None; 
  opts,pargs = getopt.getopt(sys.argv[1:],'',['h','v','vv','indir=',
  'aidA=','aidB=','infile=' ])
  if not opts: ErrorExit(usage)
  for (opt,val) in opts:
    if opt=='--h': ErrorExit(usage)
    elif opt=='--indir': indir=val
    elif opt=='--aidA': aidA=val
    elif opt=='--aidB': aidB=val
    elif opt=='--vv': verbose=2
    elif opt=='--v': verbose=1
    else: ErrorExit('Illegal option: %s'%val)

  if not (aidA and aidB):
    ErrorExit('--aidA and --aidB required\n'+usage)

  try:
    aidA=int(aidA)
    aidB=int(aidB)
  except:
    ErrorExit('aidA and aidB must be integers.\n'+usage)

###

  fpath_csv_gzA=None; fpath_csv_gzB=None;
  for fname_zip in os.listdir(indir):
    if not re.search('\.zip',fname_zip): continue
    fpath_zip=indir+'/'+fname_zip
    try:
      zf=zipfile.ZipFile(fpath_zip,'r')
    except:
      print >>sys.stderr, 'ERROR: cannot read fpath_zip: "%s"'%fpath_zip
      continue
    flist_csv_gz=zf.namelist()
    zf.close()
    for fpath_csv_gz in flist_csv_gz:
      if not re.search('\.csv\.gz',fpath_csv_gz): continue
      try:
        if re.search(r'/',fpath_csv_gz):
          txt=re.sub(r'^.*/(\d*)\.csv\.gz',r'\1',fpath_csv_gz)
        else:
          txt=re.sub(r'\.csv\.gz','',fpath_csv_gz)
        aid=int(txt)
      except:
        print >>sys.stderr, 'cannot parse AID: "%s"'%fpath_csv_gz
        print >>sys.stderr, 'DEBUG txt: "%s"'%txt
        continue
      if aid==aidA:
        fpath_zipA=fpath_zip
        fpath_csv_gzA=fpath_csv_gz
        if verbose>1:
          print >>sys.stderr, '\tA: %s: %s (%s)'%(aidA,fpath_zipA,fpath_csv_gzA)
      if aid==aidB:
        fpath_zipB=fpath_zip
        fpath_csv_gzB=fpath_csv_gz
        if verbose>1:
          print >>sys.stderr, '\tB: %s: %s (%s)'%(aidB,fpath_zipB,fpath_csv_gzB)
      if fpath_csv_gzA and fpath_csv_gzB: break
    if fpath_csv_gzA and fpath_csv_gzB: break

  if not fpath_csv_gzA:
    ErrorExit('ERROR: could not find file for AID %s'%(aidA))
  if not fpath_csv_gzB:
    ErrorExit('ERROR: could not find file for AID %s'%(aidB))


  if not os.access(SCRATCHDIR,os.R_OK):
    try:
      os.mkdir(SCRATCHDIR)
    except:
      print >>sys.stderr, 'ERROR: failed to create SCRATCHDIR %s'%SCRATCHDIR
      sys.exit(1)


  zfA=zipfile.ZipFile(fpath_zipA,'r')
  cwd=os.getcwd()
  os.chdir(SCRATCHDIR)
  zfA.extract(fpath_csv_gzA)
  os.chdir(cwd)
  zfA.close()
  f_csvA=gzip.open(SCRATCHDIR+'/'+fpath_csv_gzA)
  csvA=f_csvA.read()
  f_csvA.close()
  if not csvA:
    ErrorExit('ERROR: file empty: AID %d: %s'%(aidA,fpath_csv_gzA))

  zfB=zipfile.ZipFile(fpath_zipB,'r')
  cwd=os.getcwd()
  os.chdir(SCRATCHDIR)
  zfB.extract(fpath_csv_gzB)
  os.chdir(cwd)
  zfB.close()
  f_csvB=gzip.open(SCRATCHDIR+'/'+fpath_csv_gzB)
  csvB=f_csvB.read()
  f_csvB.close()
  if not csvB:
    ErrorExit('ERROR: file empty: AID %d: %s'%(aidB,fpath_csv_gzB))


  sids_active=[]; sids_inactive=[]; sids_inconclusive=[]; sids_unspecified=[];
  sids_discrepant=[]; sids_tested=[];
  sidset={};
  n_datapoints_total=0
  use_cids=False;

  sidsA=pubchem.ftp.Utils.ExtractOutcomes(csvA,sidset,use_cids)

  sidsB=pubchem.ftp.Utils.ExtractOutcomes(csvB,sidset,use_cids)

  print >>sys.stderr, '\t  aidA, SIDs total: %3d'%(len(sidsA.keys()))
  print >>sys.stderr, '\t  aidB, SIDs total: %3d'%(len(sidsB.keys()))

  sids_all=(set(sidsA.keys()) | set(sidsB.keys()))
  sids_common=(set(sidsA.keys()) & set(sidsB.keys()))
  print >>sys.stderr, '\t  aidA | aidB, SIDs total: %3d'%(len(sids_all))
  print >>sys.stderr, '\t  aidA & aidB, SIDs common: %3d'%(len(sids_common))


  n_activeA=0; n_inactiveA=0; n_inconclusiveA=0; n_unspecifiedA=0; n_discrepantA=0;
  n_activeB=0; n_inactiveB=0; n_inconclusiveB=0; n_unspecifiedB=0; n_discrepantB=0;
  n_active_common=0
  for sid in sids_all:
    outcomeA=None;
    outcomeB=None;
    if sidsA.has_key(sid):
      outcomeA=sidsA[sid]['outcome']
      if outcomeA==2: n_activeA+=1
      elif outcomeA==1: n_inactiveA+=1
      elif outcomeA==3: n_inconclusiveA+=1
      elif outcomeA==4: n_unspecifiedA+=1
      elif outcomeA==5: n_discrepantA+=1
      else: print >>sys.stderr, 'ERROR: outcomeA=%d'%(sidsA[sid]['outcome'])
    if sidsB.has_key(sid):
      outcomeB=sidsB[sid]['outcome']
      if outcomeB==2: n_activeB+=1
      elif outcomeB==1: n_inactiveB+=1
      elif outcomeB==3: n_inconclusiveB+=1
      elif outcomeB==4: n_unspecifiedB+=1
      elif outcomeB==5: n_discrepantB+=1
      else: print >>sys.stderr, 'ERROR: outcomeB=%d'%(sidsB[sid]['outcome'])
    if outcomeA==2 and outcomeB==2:
      n_active_common+=1

  if verbose>1:
    print >>sys.stderr, '\tA:       active: %3d'%(n_activeA)
    print >>sys.stderr, '\tA:     inactive: %3d'%(n_inactiveA)
    print >>sys.stderr, '\tA: inconclusive: %3d'%(n_inconclusiveA)
    print >>sys.stderr, '\tA:  unspecified: %3d'%(n_unspecifiedA)
    print >>sys.stderr, '\tA:   discrepant: %3d'%(n_discrepantA)
    print >>sys.stderr, '\tA:        total: %3d'%(len(sidsA.keys()))
    print >>sys.stderr, '\tB:       active: %3d'%(n_activeB)
    print >>sys.stderr, '\tB:     inactive: %3d'%(n_inactiveB)
    print >>sys.stderr, '\tB: inconclusive: %3d'%(n_inconclusiveB)
    print >>sys.stderr, '\tB:  unspecified: %3d'%(n_unspecifiedB)
    print >>sys.stderr, '\tB:   discrepant: %3d'%(n_discrepantB)
    print >>sys.stderr, '\tB:        total: %3d'%(len(sidsB.keys()))
    print >>sys.stderr, '\t  common active: %3d'%(n_active_common)

  sim = float(n_active_common) / len(sids_common)
  print >>sys.stderr, '\t  Tanimoto similarity: %.3f'%(sim)



#!/usr/bin/env python3
"""
Client for BioGRID REST API.

See: http://wiki.thebiogrid.org/doku.php/biogridrest

Formats: tab1, tab2, extendedTab2, count, json, jsonExtended

EXPERIMENTAL_SYSTEM
	"Affinity Capture-Luminescence"
	"Affinity Capture-MS"
	"Affinity Capture-Western"
	"Biochemical Activity"
	"Co-fractionation"
	"Co-localization"
	"FRET"
	"PCA"
	"Phenotypic Enhancement"
	"Phenotypic Suppression"
	"Protein-peptide"
	"Reconstituted Complex"
	"Synthetic Growth Defect"
	"Two-hybrid"
"""
###
import sys,os,re,json,argparse,time,yaml,logging
#
from .. import biogrid
#
API_HOST='webservice.thebiogrid.org'
API_BASE_PATH=''
#
##############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='BioGrid REST API client', epilog='')
  ops = ['list_organisms', 'list_idtypes', 'get_interactions', 'search_interactions']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--ids", help="query IDs (comma-separated)")
  parser.add_argument("--i", dest="ifile", help="input query IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--inc_interspecies", type=bool, help="include inter-species interactions")
  parser.add_argument("--inc_self", type=bool, help="include self-interactions")
  parser.add_argument("--inc_evidence", type=bool, help="include only ELIST evidence (else exclude)")
  parser.add_argument("--elist", default="", help="evidence codes (|-separated)")
  parser.add_argument("--addl_idtypes", default="", help="additional ID types (|-separated)")
  parser.add_argument("--human", type=bool, help="human-only")
  parser.add_argument("--rex")
  parser.add_argument("--str_query")
  parser.add_argument("--nchunk", type=int)
  parser.add_argument("--nmax", type=int, default=None)
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("--api_key", help="has precedence over param_file")
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.biogrid.yaml")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  params = biogrid.ReadParamFile(args.param_file)
  if args.api_key: params['API_KEY'] = args.api_key

  search_params = {
	'inc_interspecies': args.inc_interspecies,
	'inc_self': args.inc_self,
	'inc_evidence': args.inc_evidence,
	'elist': re.split('[|\s]+', args.elist.strip()),
	'addl_idtypes': re.split('[|\s]+', args.addl_idtypes.strip()),
	'human': args.human
	}

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.strip())
  elif args.ids:
    ids = re.split('[,\s]+', args.ids.strip())

  t0=time.time()

  if args.op =="list_organisms":
    biogrid.ListOrganisms(base_url, params, fout)

  elif args.op =="list_idtypes":
    biogrid.ListIdTypes(base_url, params, fout)

  elif args.op =="get_interactions":
    biogrid.GetInteractions(base_url, params, ids, fout)

  elif args.op =="search_interactions":
    biogrid.SearchInteractions(base_url, params, ids, search_params, fout)

  else:
    parser.error("Invalid operation: {0}".format(args.op))

  logging.info(('Elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
"""
Sparql endpoint client
"""
#############################################################################
import os,sys,json,argparse,re,time,logging

import SPARQLWrapper

from .. import sparql

FMTS={
	'JSON':SPARQLWrapper.Wrapper.JSON,
	'JSONLD':SPARQLWrapper.Wrapper.JSONLD,
	'XML':SPARQLWrapper.Wrapper.XML,
	'RDF':SPARQLWrapper.Wrapper.RDF,
	'RDFXML':SPARQLWrapper.Wrapper.RDFXML,
	'N3':SPARQLWrapper.Wrapper.N3,
	'TTL':SPARQLWrapper.Wrapper.TURTLE,
	'CSV':SPARQLWrapper.Wrapper.CSV,
	'TSV':SPARQLWrapper.Wrapper.TSV
	}

#############################################################################
if __name__=='__main__':
  epilog="SPARQLWRAPPERVERSION: {}".format(SPARQLWrapper.__version__)
  parser = argparse.ArgumentParser(description='Sparql utilities', epilog=epilog)
  endpoint="http://dbpedia.org/sparql"
  parser.add_argument("--rqfile", help="Sparql input file")
  parser.add_argument("--rq", help="Sparql input string")
  parser.add_argument("--o", dest="ofile", help="output results (TSV)")
  parser.add_argument("--endpoint", default=endpoint, help="Sparql endpoint")
  parser.add_argument("--defgraph", help="default graph URL")
  parser.add_argument("--nmax", type=int, default=1000, help="max returned triples")
  parser.add_argument("--fmt", choices=FMTS.keys(), default='JSON', help="output format")
  parser.add_argument("--test", action="store_true", help="test query")
  parser.add_argument("--test_drugname", default="metformin", help="test drugname query")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  if args.test:
    sparql.Utils.Test(args.test_drugname, FMTS[args.fmt])
    sys.exit()

  if args.rqfile:
    with open(args.rqfile) as fin:
      rqcode = fin.read()
  elif args.rq:
    rqcode = args.rq
  else:
    parser.error("--rq or --rqfile required.")

  fout = open(args.ofile,'w+') if args.ofile else sys.stdout

  t0=time.time()

  logging.info('fmt: %s'%(args.fmt))

  if args.fmt=='XML':
    results = sparql.Utils.SparqlRequest(rqcode, args.endpoint, args.defgraph, FMTS[args.fmt])
    results = results.convert()
    fout.write('%s\n'%results.toxml())

  elif args.fmt=='JSON':
    results = sparql.Utils.SparqlRequest(rqcode, args.endpoint, args.defgraph, FMTS[args.fmt])
    results = results.convert()
    fout.write('%s\n'%json.dumps(results, indent=2))

  elif args.fmt=='N3':
    results = sparql.Utils.SparqlRequest(rqcode, args.endpoint, args.defgraph, FMTS[args.fmt])
    fout.write('%s\n'%results._convertN3())

  elif args.fmt=='TSV':
    results = sparql.Utils.SparqlRequest(rqcode, args.endpoint, args.defgraph, FMTS['JSON'])
    results = results.convert()
    sparql.Utils.Results2TSV(results, [], nmax, fout)

  else:
    results = sparql.Utils.SparqlRequest(rqcode, args.endpoint, args.defgraph, None)
    fout.write('%s\n'%str(results))


